@@ -1,11 +1,11 @@
 /**
  *	@file    cryptolib.h
- *	@version 33ef80f (HEAD, tag: MATRIXSSL-3-7-2-OPEN, tag: MATRIXSSL-3-7-2-COMM, origin/master, origin/HEAD, master)
+ *	@version 7f5c362 (HEAD -> master)
  *
  *	Header file for definitions used with crypto lib.
  */
 /*
- *	Copyright (c) 2013-2015 INSIDE Secure Corporation
+ *	Copyright (c) 2013-2016 INSIDE Secure Corporation
  *	Copyright (c) PeerSec Networks, 2002-2011
  *	All Rights Reserved
  *
@@ -19,7 +19,7 @@
  *	This General Public License does NOT permit incorporating this software
  *	into proprietary programs.  If you are unable to comply with the GPL, a
  *	commercial license for this software may be purchased from INSIDE at
- *	http://www.insidesecure.com/eng/Company/Locations
+ *	http://www.insidesecure.com/
  *
  *	This program is distributed in WITHOUT ANY WARRANTY; without even the
  *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
@@ -35,74 +35,46 @@
 #ifndef _h_PS_CRYPTOLIB
 #define _h_PS_CRYPTOLIB
 
-#ifdef USE_PKCS11
-#include "cryptoki.h" /* #include "pkcs11.h" */
-#endif
-#ifdef USE_AESNI_CRYPTO
-#include "hardware/aesni.h"
-#endif
-#include "symmetric/symmetric.h"
-#include "math/pstm.h"
+/******************************************************************************/
+/*
+	Additional 'hidden' algorithm configuration here for deprecated support
+*/
 
+/** Symmetric. @security These are generally insecure and not enabled by default. */
+//#define USE_ARC4
+//#define USE_SEED
+//#define USE_IDEA
+#ifdef USE_PKCS12
+//#define USE_RC2	/* Only PKCS#12 parse should ever want this algorithm */
+#endif
+
+/** Digest. @security These are generally insecure and not enabled by default */
+//#define USE_MD4
+//#define USE_MD2
+
+/** PRNG. @security By default the OS PRNG will be used directly. */
+#define USE_PRNG
+//#define USE_YARROW
 
 /******************************************************************************/
 /*
-	Configuration checks
+	Additional configuration that is usually not modified.
 */
-#ifdef USE_CERT_PARSE
-	#ifndef USE_X509
-	#error "Must enable USE_X509 if USE_CERT_PARSE is enabled"
-	#endif
-	#if !defined(USE_MD5) || !defined(USE_SHA1)
-	#error "Both USE_MD5 and USE_SHA1 must be enabled when enabling USE_X509"
-	#endif
-#endif
+#define OCSP_VALID_TIME_WINDOW 604800 /* In seconds (1 week default window) */
 
-#ifdef USE_HMAC
-	#if !defined(USE_SHA1) && !defined(USE_SHA256) && !defined(USE_SHA384)
-	#error "Must enable a SHA based hash in cryptoConfig.h for HMAC support"
-	#endif
-#endif
+/******************************************************************************/
+/*
+	Include crypto provider layer headers
+*/
+#include "layer/layer.h"
 
-#ifdef USE_PKCS5
-	#ifndef USE_MD5
-	#error "Enable USE_MD5 in cryptoConfig.h for PKCS5 support"
-	#endif
-	#ifndef USE_3DES
-	#error "Enable USE_3DES in cryptoConfig.h for PKCS5 support"
-	#endif
-	#ifndef USE_AES
-	#error "Enable USE_AES in cryptoConfig.h for PKCS5 support"
-	#endif
-#endif
-
-#ifdef USE_PKCS8
-	#ifndef USE_HMAC
-	#error "Enable USE_HMAC in cryptoConfig.h for PKCS8 support"
-	#endif
-#endif
-
-#ifdef USE_PKCS11
-	#ifdef USE_RSA
-	#error "Currently no RSA support for PKCS#11 builds.  Disable USE_RSA"
-	#endif
-	#define USE_UNIFIED_PKCS11 /* TLS integration is mandatory right now */
-	#define USE_PKCS11_ECC
-	#define USE_PKCS11_AES
-	#define USE_PKCS11_HASH
-#else
-	#define USE_NATIVE_ECC
-	#define USE_NATIVE_AES
-	#define USE_NATIVE_HASH
-#endif
-
-#ifdef USE_PKCS12
-	#ifndef USE_PKCS8
-	#error "Enable USE_PKCS8 in cryptoConfig.h for PKCS12 support"
-	#endif
-#endif
+/* Configuration validation/sanity checks */
+#include "cryptoCheck.h"
 
+/* Implementation layer */
+#include "symmetric/symmetric.h"
 #include "digest/digest.h"
+#include "math/pstm.h"
 #include "pubkey/pubkey.h"
 #include "keyformat/asn1.h"
 #include "keyformat/x509.h"
@@ -124,16 +96,15 @@
 #define psTracePtrCrypto(x, y) _psTracePtr(x, y)
 #endif /* USE_CRYPTO_TRACE */
 
-
 /******************************************************************************/
 /*
 	Helpers
 */
-extern int32 psBase64decode(const unsigned char *in,  uint32 len,
-					unsigned char *out, uint32 *outlen);
+extern int32_t psBase64decode(const unsigned char *in, uint16_t len,
+					unsigned char *out, uint16_t *outlen);
 extern void psOpenPrng(void);
 extern void psClosePrng(void);
-extern int32 matrixCryptoGetPrngData(unsigned char *bytes, uint32 size,
+extern int32_t matrixCryptoGetPrngData(unsigned char *bytes, uint16_t size,
 					void *userPtr);
 
 /******************************************************************************/
@@ -245,29 +216,32 @@ extern int32 matrixCryptoGetPrngData(unsigned char *bytes, uint32 size,
 
 /******************************************************************************/
 /* These values are all mutually exlusive bits to define Cipher flags */
-#define CRYPTO_FLAGS_AES	0x01
-#define CRYPTO_FLAGS_AES256	0x02
-#define CRYPTO_FLAGS_3DES	0x04
-#define CRYPTO_FLAGS_ARC4	0x08
-#define CRYPTO_FLAGS_SEED	0x10
+#define CRYPTO_FLAGS_AES	(1<<0)
+#define CRYPTO_FLAGS_AES256	(1<<1)
+#define CRYPTO_FLAGS_3DES	(1<<2)
+#define CRYPTO_FLAGS_ARC4	(1<<3)
+#define CRYPTO_FLAGS_SEED	(1<<4)
+#define CRYPTO_FLAGS_IDEA	(1<<5)
+#define CRYPTO_FLAGS_CHACHA (1<<6)	/* Short for CHACHA20_POLY2305 */
 
-#define CRYPTO_FLAGS_SHA1	0x20
-#define CRYPTO_FLAGS_SHA2	0x40
-#define CRYPTO_FLAGS_MD5	0x80
+#define CRYPTO_FLAGS_SHA1	(1<< 8)
+#define CRYPTO_FLAGS_SHA2	(1<< 9)
+#define CRYPTO_FLAGS_SHA3	(1<<10)
+#define CRYPTO_FLAGS_GCM	(1<<11)
+#define CRYPTO_FLAGS_CCM	(1<<12)
+#define CRYPTO_FLAGS_CCM8	(1<<13)	/* CCM mode with 8 byte ICV */
+#define CRYPTO_FLAGS_MD5	(1<<14)
 
-#define CRYPTO_FLAGS_TLS		0x100
-#define CRYPTO_FLAGS_TLS_1_1	0x200
-#define CRYPTO_FLAGS_TLS_1_2	0x400
+#define CRYPTO_FLAGS_TLS		(1<<16)
+#define CRYPTO_FLAGS_TLS_1_1	(1<<17)
+#define CRYPTO_FLAGS_TLS_1_2	(1<<18)
 
-#define CRYPTO_FLAGS_INBOUND	0x800
-#define CRYPTO_FLAGS_ARC4INITE	0x1000 /* Encrypt init */
-#define CRYPTO_FLAGS_ARC4INITD	0x2000 /* Decrypt init */
-#define CRYPTO_FLAGS_BLOCKING	0x4000
+#define CRYPTO_FLAGS_INBOUND	(1<<24)
+#define CRYPTO_FLAGS_ARC4INITE	(1<<25)
+#define CRYPTO_FLAGS_ARC4INITD	(1<<26)
+#define CRYPTO_FLAGS_BLOCKING	(1<<27)
 
-#define CRYPTO_FLAGS_DISABLED	0x8000
-#define CRYPTO_FLAGS_GCM		0x10000
-
-#define CRYPTO_FLAGS_SHA3		0x20000 /* SHA-384 */
+#define CRYPTO_FLAGS_DISABLED	(1<<30)
 
 /******************************************************************************/
 
@@ -290,7 +264,7 @@ extern int32 matrixCryptoGetPrngData(unsigned char *bytes, uint32 size,
 #elif defined(__GNUC__) && (defined(__i386__) || defined(__x86_64__)) && \
 		!defined(INTEL_CC) && !defined(PS_NO_ASM)
 
-static inline unsigned ROL(unsigned word, int i)
+static __inline unsigned ROL(unsigned word, int i)
 {
    asm ("roll %%cl,%0"
 	  :"=r" (word)
@@ -298,7 +272,7 @@ static inline unsigned ROL(unsigned word, int i)
    return word;
 }
 
-static inline unsigned ROR(unsigned word, int i)
+static __inline unsigned ROR(unsigned word, int i)
 {
    asm ("rorl %%cl,%0"
 	  :"=r" (word)
@@ -459,7 +433,7 @@ x = (((uint64)((y)[7] & 255))<<56)|(((uint64)((y)[6] & 255))<<48)| \
 
 #else /* 64-bit words then  */
 #define STORE32L(x, y) \
-{ unsigned long __t = (x); memcpy(y, &__t, 4); }
+{ unsigned int __t = (x); memcpy(y, &__t, 4); }
 
 #define LOAD32L(x, y) \
 { memcpy(&(x), y, 4); x &= 0xFFFFFFFF; }
@@ -473,6 +447,8 @@ x = (((uint64)((y)[7] & 255))<<56)|(((uint64)((y)[6] & 255))<<48)| \
 #endif /* ENDIAN_64BITWORD */
 #endif /* ENDIAN_LITTLE */
 
+/******************************************************************************/
+
 #ifdef ENDIAN_BIG
 #define STORE32L(x, y) { \
 (y)[3] = (unsigned char)(((x)>>24)&255); \
@@ -506,6 +482,8 @@ x = (((uint64)((y)[7] & 255))<<56)|(((uint64)((y)[6] & 255))<<48) | \
 (((uint64)((y)[1] & 255))<<8)|(((uint64)((y)[0] & 255))); \
 }
 
+/******************************************************************************/
+
 #ifdef ENDIAN_32BITWORD
 #define STORE32H(x, y) \
 { unsigned int __t = (x); memcpy(y, &__t, 4); }
@@ -530,10 +508,12 @@ x = (((uint64)((y)[0] & 255))<<56)|(((uint64)((y)[1] & 255))<<48)| \
 (((uint64)((y)[6] & 255))<<8)| (((uint64)((y)[7] & 255))); \
 }
 
+/******************************************************************************/
+
 #else /* 64-bit words then  */
 
 #define STORE32H(x, y) \
-{ unsigned long __t = (x); memcpy(y, &__t, 4); }
+{ unsigned int __t = (x); memcpy(y, &__t, 4); }
 
 #define LOAD32H(x, y) \
 { memcpy(&(x), y, 4); x &= 0xFFFFFFFF; }
@@ -547,6 +527,8 @@ x = (((uint64)((y)[0] & 255))<<56)|(((uint64)((y)[1] & 255))<<48)| \
 #endif /* ENDIAN_64BITWORD */
 #endif /* ENDIAN_BIG */
 
+/******************************************************************************/
+
 #ifdef HAVE_NATIVE_INT64
 #define ROL64c(x, y) \
 ( (((x)<<((uint64)(y)&63)) | \
@@ -556,8 +538,6 @@ x = (((uint64)((y)[0] & 255))<<56)|(((uint64)((y)[1] & 255))<<48)| \
 ( ((((x)&CONST64(0xFFFFFFFFFFFFFFFF))>>((uint64)(y)&CONST64(63))) | \
 ((x)<<((uint64)(64-((y)&CONST64(63)))))) & CONST64(0xFFFFFFFFFFFFFFFF))
 #endif /* HAVE_NATIVE_INT64 */
-/******************************************************************************/
-
 
 /******************************************************************************/
 /*
@@ -571,26 +551,12 @@ x = (((uint64)((y)[0] & 255))<<56)|(((uint64)((y)[1] & 255))<<48)| \
 	BLOCKSIZE <= 1 ? (unsigned char)0 : \
 	(unsigned char)(BLOCKSIZE - ((LEN) & (BLOCKSIZE - 1)))
 
-
-#ifdef USE_PKCS11
-
-/* #define PKCS11_STATS */
-#ifdef PKCS11_STATS
-extern void pkcs11RegisterObj(CK_SESSION_HANDLE ses, CK_OBJECT_HANDLE obj);
-extern void pkcs11ShowObjects();
+#ifdef  USE_CRL
+extern int32_t psCrlOpen();
+extern void psCrlClose();
 #endif
-extern CK_RV pkcs11Init(CK_C_INITIALIZE_ARGS *args);
-extern CK_RV pkcs11Close(void);
-extern CK_RV pkcs11OpenSession(CK_SESSION_HANDLE *session, int32 flags);
-extern void pkcs11CloseSession(CK_SESSION_HANDLE session);
-extern CK_RV pkcs11CreateObject(CK_SESSION_HANDLE session,
-				CK_ATTRIBUTE *pTemplate, CK_ULONG ulCount,
-				CK_OBJECT_HANDLE *phObject);
-extern CK_RV pkcs11DestroyObject(CK_SESSION_HANDLE session,
-				CK_OBJECT_HANDLE object);
-#endif
-
 
 #endif /* _h_PS_CRYPTOLIB */
+
 /******************************************************************************/
 
