@@ -1,11 +1,11 @@
 /**
  *	@file    pstm_mul_comba.c
- *	@version 33ef80f (HEAD, tag: MATRIXSSL-3-7-2-OPEN, tag: MATRIXSSL-3-7-2-COMM, origin/master, origin/HEAD, master)
+ *	@version 7f5c362 (HEAD -> master)
  *
  *	Multiprecision multiplication with Comba technique.
  */
 /*
- *	Copyright (c) 2013-2015 INSIDE Secure Corporation
+ *	Copyright (c) 2013-2016 INSIDE Secure Corporation
  *	Copyright (c) PeerSec Networks, 2002-2011
  *	All Rights Reserved
  *
@@ -19,7 +19,7 @@
  *	This General Public License does NOT permit incorporating this software
  *	into proprietary programs.  If you are unable to comply with the GPL, a
  *	commercial license for this software may be purchased from INSIDE at
- *	http://www.insidesecure.com/eng/Company/Locations
+ *	http://www.insidesecure.com/
  *
  *	This program is distributed in WITHOUT ANY WARRANTY; without even the
  *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
@@ -33,7 +33,8 @@
 /******************************************************************************/
 
 #include "../cryptoApi.h"
-#ifndef DISABLE_PSTM
+
+#if defined(USE_MATRIX_RSA) || defined(USE_MATRIX_ECC) || defined(USE_MATRIX_DH)
 
 /******************************************************************************/
 #if defined(PSTM_X86)
@@ -73,7 +74,7 @@ asm(                                                      \
 	 "addl  %%eax,%0     \n\t"                            \
 	 "adcl  %%edx,%1     \n\t"                            \
 	 "adcl  $0,%2        \n\t"                            \
-	 :"=r"(c0), "=r"(c1), "=r"(c2): "0"(c0), "1"(c1), "2"(c2), "m"(i), "m"(j)  :"%eax","%edx","%cc");
+	 :"=r"(c0), "=r"(c1), "=r"(c2): "0"(c0), "1"(c1), "2"(c2), "m"(i), "m"(j)  :"%eax","%edx","cc");
 
 /******************************************************************************/
 #elif defined(PSTM_X86_64)
@@ -142,7 +143,7 @@ asm(                                                          \
 	"  ADDS   %0,%0,r0              \n\t"                     \
 	"  ADCS   %1,%1,r1              \n\t"                     \
 	"  ADC    %2,%2,#0              \n\t"                     \
-	:"=r"(c0), "=r"(c1), "=r"(c2) : "0"(c0), "1"(c1), "2"(c2), "r"(i), "r"(j) : "r0", "r1", "%cc");
+	:"=r"(c0), "=r"(c1), "=r"(c2) : "0"(c0), "1"(c1), "2"(c2), "r"(i), "r"(j) : "r0", "r1", "cc");
 
 /******************************************************************************/
 #elif defined(PSTM_MIPS)
@@ -210,8 +211,9 @@ asm(                               \
 
 /******************************************************************************/
 /* generic PxQ multiplier */
-static int32 pstm_mul_comba_gen(psPool_t *pool, pstm_int *A, pstm_int *B,
-			pstm_int *C, pstm_digit *paD, uint32 paDlen)
+static int32_t pstm_mul_comba_gen(psPool_t *pool, const pstm_int *A,
+				const pstm_int *B, pstm_int *C,
+				pstm_digit *paD, uint16_t paDlen)
 {
 	int16		paDfail, pa;
 	int32       ix, iy, iz, tx, ty;
@@ -304,7 +306,7 @@ static int32 pstm_mul_comba_gen(psPool_t *pool, pstm_int *A, pstm_int *B,
 
 /******************************************************************************/
 #ifdef USE_1024_KEY_SPEED_OPTIMIZATIONS
-static int32 pstm_mul_comba16(pstm_int *A, pstm_int *B, pstm_int *C)
+static int32_t pstm_mul_comba16(const pstm_int *A, const pstm_int *B, pstm_int *C)
 {
 	pstm_digit c0, c1, c2, at[32];
 
@@ -453,7 +455,7 @@ static int32 pstm_mul_comba16(pstm_int *A, pstm_int *B, pstm_int *C)
 
 
 #ifdef USE_2048_KEY_SPEED_OPTIMIZATIONS
-static int32 pstm_mul_comba32(pstm_int *A, pstm_int *B, pstm_int *C)
+static int32_t pstm_mul_comba32(const pstm_int *A, const pstm_int *B, pstm_int *C)
 {
    pstm_digit c0, c1, c2, at[64];
    int32 out_size;
@@ -744,8 +746,8 @@ static int32 pstm_mul_comba32(pstm_int *A, pstm_int *B, pstm_int *C)
 
 /******************************************************************************/
 
-int32 pstm_mul_comba(psPool_t *pool, pstm_int *A, pstm_int *B, pstm_int *C,
-			pstm_digit *paD, uint32 paDlen)
+int32 pstm_mul_comba(psPool_t *pool, const pstm_int *A, const pstm_int *B, pstm_int *C,
+			pstm_digit *paD, uint16_t paDlen)
 {
 #ifdef USE_1024_KEY_SPEED_OPTIMIZATIONS
 	if (A->used == 16 && B->used == 16) {
@@ -768,5 +770,7 @@ int32 pstm_mul_comba(psPool_t *pool, pstm_int *A, pstm_int *B, pstm_int *C,
 #endif
 }
 
-#endif /* !DISABLE_PSTM */
+#endif /* defined(USE_MATRIX_RSA) || defined(USE_MATRIX_ECC) */
+
 /******************************************************************************/
+
