@@ -1,11 +1,11 @@
 /**
  *	@file    pstm_sqr_comba.c
- *	@version 33ef80f (HEAD, tag: MATRIXSSL-3-7-2-OPEN, tag: MATRIXSSL-3-7-2-COMM, origin/master, origin/HEAD, master)
+ *	@version 7f5c362 (HEAD -> master)
  *
  *	Multiprecision Squaring with Comba technique.
  */
 /*
- *	Copyright (c) 2013-2015 INSIDE Secure Corporation
+ *	Copyright (c) 2013-2016 INSIDE Secure Corporation
  *	Copyright (c) PeerSec Networks, 2002-2011
  *	All Rights Reserved
  *
@@ -19,7 +19,7 @@
  *	This General Public License does NOT permit incorporating this software
  *	into proprietary programs.  If you are unable to comply with the GPL, a
  *	commercial license for this software may be purchased from INSIDE at
- *	http://www.insidesecure.com/eng/Company/Locations
+ *	http://www.insidesecure.com/
  *
  *	This program is distributed in WITHOUT ANY WARRANTY; without even the
  *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
@@ -33,7 +33,8 @@
 /******************************************************************************/
 
 #include "../cryptoApi.h"
-#ifndef DISABLE_PSTM
+
+#if defined(USE_MATRIX_RSA) || defined(USE_MATRIX_ECC) || defined(USE_MATRIX_DH)
 
 /******************************************************************************/
 #if defined(PSTM_X86)
@@ -66,7 +67,7 @@ asm(                                            \
 	 "addl  %%eax,%0     \n\t"                            \
 	 "adcl  %%edx,%1     \n\t"                            \
 	 "adcl  $0,%2        \n\t"                            \
-	 :"=r"(c0), "=r"(c1), "=r"(c2): "0"(c0), "1"(c1), "2"(c2), "m"(i) :"%eax","%edx","%cc");
+	 :"=r"(c0), "=r"(c1), "=r"(c2): "0"(c0), "1"(c1), "2"(c2), "m"(i) :"%eax","%edx","cc");
 
 #define SQRADD2(i, j)                                     \
 asm(                                            \
@@ -78,7 +79,7 @@ asm(                                            \
 	 "addl  %%eax,%0     \n\t"                            \
 	 "adcl  %%edx,%1     \n\t"                            \
 	 "adcl  $0,%2        \n\t"                            \
-	 :"=r"(c0), "=r"(c1), "=r"(c2): "0"(c0), "1"(c1), "2"(c2), "m"(i), "m"(j)  :"%eax","%edx","%cc");
+	 :"=r"(c0), "=r"(c1), "=r"(c2): "0"(c0), "1"(c1), "2"(c2), "m"(i), "m"(j)  :"%eax","%edx","cc");
 
 #define SQRADDSC(i, j)                                    \
 asm(                                                     \
@@ -87,7 +88,7 @@ asm(                                                     \
 	 "movl  %%eax,%0     \n\t"                            \
 	 "movl  %%edx,%1     \n\t"                            \
 	 "xorl  %2,%2        \n\t"                            \
-	 :"=r"(sc0), "=r"(sc1), "=r"(sc2): "0"(sc0), "1"(sc1), "2"(sc2), "g"(i), "g"(j) :"%eax","%edx","%cc");
+	 :"=r"(sc0), "=r"(sc1), "=r"(sc2): "0"(sc0), "1"(sc1), "2"(sc2), "g"(i), "g"(j) :"%eax","%edx","cc");
 
 #define SQRADDAC(i, j)                                    \
 asm(                                                     \
@@ -96,7 +97,7 @@ asm(                                                     \
 	 "addl  %%eax,%0     \n\t"                            \
 	 "adcl  %%edx,%1     \n\t"                            \
 	 "adcl  $0,%2        \n\t"                            \
-	 :"=r"(sc0), "=r"(sc1), "=r"(sc2): "0"(sc0), "1"(sc1), "2"(sc2), "g"(i), "g"(j) :"%eax","%edx","%cc");
+	 :"=r"(sc0), "=r"(sc1), "=r"(sc2): "0"(sc0), "1"(sc1), "2"(sc2), "g"(i), "g"(j) :"%eax","%edx","cc");
 
 #define SQRADDDB                                          \
 asm(                                                     \
@@ -106,7 +107,7 @@ asm(                                                     \
 	 "addl %6,%0         \n\t"                            \
 	 "adcl %7,%1         \n\t"                            \
 	 "adcl %8,%2         \n\t"                            \
-	 :"=r"(c0), "=r"(c1), "=r"(c2) : "0"(c0), "1"(c1), "2"(c2), "r"(sc0), "r"(sc1), "r"(sc2) : "%cc");
+	 :"=r"(c0), "=r"(c1), "=r"(c2) : "0"(c0), "1"(c1), "2"(c2), "r"(sc0), "r"(sc1), "r"(sc2) : "cc");
 
 /******************************************************************************/
 #elif defined(PSTM_X86_64)
@@ -209,7 +210,7 @@ asm(                                                             \
 "  ADDS   %0,%0,r0                 \n\t"                         \
 "  ADCS   %1,%1,r1                 \n\t"                         \
 "  ADC    %2,%2,#0                 \n\t"                         \
-:"=r"(c0), "=r"(c1), "=r"(c2) : "0"(c0), "1"(c1), "2"(c2), "r"(i) : "r0", "r1", "%cc");
+:"=r"(c0), "=r"(c1), "=r"(c2) : "0"(c0), "1"(c1), "2"(c2), "r"(i) : "r0", "r1", "cc");
 
 /* for squaring some of the terms are doubled... */
 #define SQRADD2(i, j)                                            \
@@ -221,13 +222,13 @@ asm(                                                             \
 "  ADDS   %0,%0,r0                 \n\t"                         \
 "  ADCS   %1,%1,r1                 \n\t"                         \
 "  ADC    %2,%2,#0                 \n\t"                         \
-:"=r"(c0), "=r"(c1), "=r"(c2) : "0"(c0), "1"(c1), "2"(c2), "r"(i), "r"(j) : "r0", "r1", "%cc");
+:"=r"(c0), "=r"(c1), "=r"(c2) : "0"(c0), "1"(c1), "2"(c2), "r"(i), "r"(j) : "r0", "r1", "cc");
 
 #define SQRADDSC(i, j)                                           \
 asm(                                                             \
 "  UMULL  %0,%1,%6,%7              \n\t"                         \
 "  SUB    %2,%2,%2                 \n\t"                         \
-:"=r"(sc0), "=r"(sc1), "=r"(sc2) : "0"(sc0), "1"(sc1), "2"(sc2), "r"(i), "r"(j) : "%cc");
+:"=r"(sc0), "=r"(sc1), "=r"(sc2) : "0"(sc0), "1"(sc1), "2"(sc2), "r"(i), "r"(j) : "cc");
 
 #define SQRADDAC(i, j)                                           \
 asm(                                                             \
@@ -235,7 +236,7 @@ asm(                                                             \
 "  ADDS   %0,%0,r0                 \n\t"                         \
 "  ADCS   %1,%1,r1                 \n\t"                         \
 "  ADC    %2,%2,#0                 \n\t"                         \
-:"=r"(sc0), "=r"(sc1), "=r"(sc2) : "0"(sc0), "1"(sc1), "2"(sc2), "r"(i), "r"(j) : "r0", "r1", "%cc");
+:"=r"(sc0), "=r"(sc1), "=r"(sc2) : "0"(sc0), "1"(sc1), "2"(sc2), "r"(i), "r"(j) : "r0", "r1", "cc");
 
 #define SQRADDDB                                                 \
 asm(                                                             \
@@ -245,7 +246,7 @@ asm(                                                             \
 "  ADDS  %0,%0,%3                     \n\t"                      \
 "  ADCS  %1,%1,%4                     \n\t"                      \
 "  ADC   %2,%2,%5                     \n\t"                      \
-:"=r"(c0), "=r"(c1), "=r"(c2) : "r"(sc0), "r"(sc1), "r"(sc2), "0"(c0), "1"(c1), "2"(c2) : "%cc");
+:"=r"(c0), "=r"(c1), "=r"(c2) : "r"(sc0), "r"(sc1), "r"(sc2), "0"(c0), "1"(c1), "2"(c2) : "cc");
 
 /******************************************************************************/
 #elif defined(PSTM_MIPS)
@@ -316,7 +317,7 @@ asm(                               \
 	" mflo   %0             \n\t"  \
 	" mfhi   %1             \n\t"  \
 	" xor    %2,%2,%2       \n\t"  \
-	:"=r"(sc0), "=r"(sc1), "=r"(sc2):"0"(sc0), "1"(sc1), "2"(sc2), "r"(i),"r"(j) : "%cc");
+	:"=r"(sc0), "=r"(sc1), "=r"(sc2):"0"(sc0), "1"(sc1), "2"(sc2), "r"(i),"r"(j) : "cc");
 
 #define SQRADDAC(i, j)            \
 asm(                              \
@@ -427,8 +428,8 @@ asm(                               \
 /*
 	Non-unrolled comba squarer
  */
-static int32 pstm_sqr_comba_gen(psPool_t *pool, pstm_int *A, pstm_int *B,
-			pstm_digit *paD, uint32 paDlen)
+static int32_t pstm_sqr_comba_gen(psPool_t *pool, const pstm_int *A,
+				pstm_int *B, pstm_digit *paD, uint16_t paDlen)
 {
 	int16		paDfail, pa;
 	int32       ix, iz;
@@ -481,15 +482,13 @@ static int32 pstm_sqr_comba_gen(psPool_t *pool, pstm_int *A, pstm_int *B,
 		/* setup temp aliases */
 		tmpx = A->dp + tx;
 		tmpy = A->dp + ty;
-
 /*
-			This is the number of times the loop will iterate,
-				while (tx++ < a->used && ty-- >= 0) { ... }
+		This is the number of times the loop will iterate
+		while (tx++ < a->used && ty-- >= 0) { ... }
 */
 		iy = min(A->used-tx, ty+1);
-
 /*
-		now for squaring tx can never equal ty. We halve the distance since
+		now for squaring, tx can never equal ty. We halve the distance since
 		they approach at a rate of 2x and we have to round because odd cases
 		need to be executed
 */
@@ -542,7 +541,7 @@ static int32 pstm_sqr_comba_gen(psPool_t *pool, pstm_int *A, pstm_int *B,
 	Unrolled Comba loop for 1024 bit keys
  */
 #ifdef USE_1024_KEY_SPEED_OPTIMIZATIONS
-static int32 pstm_sqr_comba16(pstm_int *A, pstm_int *B)
+static int32_t pstm_sqr_comba16(const pstm_int *A, pstm_int *B)
 {
 	pstm_digit *a, b[32], c0, c1, c2, sc0, sc1, sc2;
 #ifdef PSTM_ISO
@@ -728,7 +727,7 @@ static int32 pstm_sqr_comba16(pstm_int *A, pstm_int *B)
 
 
 #ifdef USE_2048_KEY_SPEED_OPTIMIZATIONS
-static int32 pstm_sqr_comba32(pstm_int *A, pstm_int *B)
+static int32_t pstm_sqr_comba32(const pstm_int *A, pstm_int *B)
 {
    pstm_digit *a, b[64], c0, c1, c2, sc0, sc1, sc2;
 #ifdef PSTM_ISO
@@ -1072,10 +1071,16 @@ static int32 pstm_sqr_comba32(pstm_int *A, pstm_int *B)
 #endif /* USE_2048_KEY_SPEED_OPTIMIZATIONS */
 
 /******************************************************************************/
-/*
- */
-int32 pstm_sqr_comba(psPool_t *pool, pstm_int *A, pstm_int *B, pstm_digit *paD,
-		uint32 paDlen)
+/**
+	B = A**2.
+	@param[in] pool Memory pool
+	@param[in] A Base
+	@param[out] B Result
+	@param[in,out] paD Temporary storage
+	@param[in] paDlen Number of items pointed to by paD
+*/
+int32_t pstm_sqr_comba(psPool_t *pool, const pstm_int *A, pstm_int *B,
+				pstm_digit *paD, uint16_t paDlen)
 {
 #ifdef USE_1024_KEY_SPEED_OPTIMIZATIONS
 	if (A->used == 16) {
@@ -1098,5 +1103,7 @@ int32 pstm_sqr_comba(psPool_t *pool, pstm_int *A, pstm_int *B, pstm_digit *paD,
 #endif
 }
 
-#endif /* DISABLE_PSTM */
+#endif /* defined(USE_MATRIX_RSA) || defined(USE_MATRIX_ECC) */
+
 /******************************************************************************/
+
