@@ -1,11 +1,11 @@
 /**
  *	@file    aes.c
- *	@version 33ef80f (HEAD, tag: MATRIXSSL-3-7-2-OPEN, tag: MATRIXSSL-3-7-2-COMM, origin/master, origin/HEAD, master)
+ *	@version 7f5c362 (HEAD -> master)
  *
- *	AES CBC block cipher implementation.
+ *	AES block cipher implementation.
  */
 /*
- *	Copyright (c) 2013-2015 INSIDE Secure Corporation
+ *	Copyright (c) 2013-2016 INSIDE Secure Corporation
  *	Copyright (c) PeerSec Networks, 2002-2011
  *	All Rights Reserved
  *
@@ -19,7 +19,7 @@
  *	This General Public License does NOT permit incorporating this software
  *	into proprietary programs.  If you are unable to comply with the GPL, a
  *	commercial license for this software may be purchased from INSIDE at
- *	http://www.insidesecure.com/eng/Company/Locations
+ *	http://www.insidesecure.com/
  *
  *	This program is distributed in WITHOUT ANY WARRANTY; without even the
  *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
@@ -34,10 +34,174 @@
 
 #include "../cryptoApi.h"
 
-#ifdef USE_AES
-#ifndef USE_AES_CBC_EXTERNAL
+#ifdef USE_MATRIX_AES_BLOCK
+
 /******************************************************************************/
 
+__inline static void invertKeySchedule(psAesKey_t *key);
+static uint32 setup_mix(uint32 temp);
+
+static const uint32 rcon[] = {
+	0x01000000UL, 0x02000000UL, 0x04000000UL, 0x08000000UL,
+	0x10000000UL, 0x20000000UL, 0x40000000UL, 0x80000000UL,
+	0x1B000000UL, 0x36000000UL
+	/* for 128-bit blocks, AES never uses more than 10 rcon values */
+};
+
+#define byte(x, n) (((x) >> (8 * (n))) & 255)
+
+/******************************************************************************/
+/**
+	Initialize the key schedule of an AES (Rijndael) block cipher.
+	The initilaized key can be used only for encrypt or decrypt.
+	If the same key bytes need to be used for both, two psAesKey_t structures
+	should be initialized.
+
+	@param[in,out] key Pointer to an allocated key structure to be initialized.
+	@param[in] ckey The symmetric key as bytes.
+	@param[in] keylen The number of bytes in 'ckey'.
+	@param[in] flags PS_AES_ENCRYPT or PS_AES_DECRYPT.
+	@return PS_SUCCESS on success, < 0 on failure.
+
+	@security psAesClearBlockKey() should be called as soon as the key is no
+	longer needed.
+	@security If 'ckey' is no longer needed after this api call, it should be
+	cleared.
+*/
+int32_t psAesInitBlockKey(psAesKey_t *key,
+				const unsigned char ckey[AES_MAXKEYLEN], uint8_t keylen,
+				uint32_t flags)
+{
+	int			i;
+	uint32		temp, *rk;
+
+#ifdef CRYPTO_ASSERT
+	if (ckey == NULL || key == NULL ||
+		!(flags & (PS_AES_ENCRYPT | PS_AES_DECRYPT))) {
+
+		psTraceCrypto("Bad args to psAesInitBlockKey\n");
+		return PS_ARG_FAIL;
+	}
+	if (keylen != AES128_KEYLEN && keylen != AES192_KEYLEN &&
+			keylen != AES256_KEYLEN) {
+		psTraceCrypto("Invalid AES key length\n");
+		return PS_ARG_FAIL;
+	}
+#endif
+	/*
+		Setup the forward (encrypt) key schedule. If flags & PS_AES_DECRYPT we
+		will calculate the inverse (decrypt) key schedule from this forward
+		schedule and replace it.
+	*/
+	i = 0;
+	rk = key->skey;
+	LOAD32H(rk[0], ckey	    );
+	LOAD32H(rk[1], ckey +  4);
+	LOAD32H(rk[2], ckey +  8);
+	LOAD32H(rk[3], ckey + 12);
+	switch (keylen) {
+	case 16:
+		key->rounds = 10;
+		for (;;) {
+			temp  = rk[3];
+			rk[4] = rk[0] ^ setup_mix(temp) ^ rcon[i];
+			rk[5] = rk[1] ^ rk[4];
+			rk[6] = rk[2] ^ rk[5];
+			rk[7] = rk[3] ^ rk[6];
+			if (++i == 10) {
+				break;
+			}
+			rk += 4;
+		}
+		break;
+
+	case 24:
+		key->rounds = 12;
+		LOAD32H(rk[4], ckey + 16);
+		LOAD32H(rk[5], ckey + 20);
+		for (;;) {
+		#ifdef _MSC_VER
+			temp = key->skey[rk - key->skey + 5];
+		#else
+			temp = rk[5];
+		#endif /* _MSC_VER */
+			rk[ 6] = rk[ 0] ^ setup_mix(temp) ^ rcon[i];
+			rk[ 7] = rk[ 1] ^ rk[ 6];
+			rk[ 8] = rk[ 2] ^ rk[ 7];
+			rk[ 9] = rk[ 3] ^ rk[ 8];
+			if (++i == 8) {
+				break;
+			}
+			rk[10] = rk[ 4] ^ rk[ 9];
+			rk[11] = rk[ 5] ^ rk[10];
+			rk += 6;
+		}
+		break;
+
+	case 32:
+		key->rounds = 14;
+		LOAD32H(rk[4], ckey + 16);
+		LOAD32H(rk[5], ckey + 20);
+		LOAD32H(rk[6], ckey + 24);
+		LOAD32H(rk[7], ckey + 28);
+		for (;;) {
+		#ifdef _MSC_VER
+			temp = key->skey[rk - key->skey + 7];
+		#else
+			temp = rk[7];
+		#endif /* _MSC_VER */
+			rk[ 8] = rk[ 0] ^ setup_mix(temp) ^ rcon[i];
+			rk[ 9] = rk[ 1] ^ rk[ 8];
+			rk[10] = rk[ 2] ^ rk[ 9];
+			rk[11] = rk[ 3] ^ rk[10];
+			if (++i == 7) {
+				break;
+			}
+			temp = rk[11];
+			rk[12] = rk[ 4] ^ setup_mix(ROR(temp, 8));
+			rk[13] = rk[ 5] ^ rk[12];
+			rk[14] = rk[ 6] ^ rk[13];
+			rk[15] = rk[ 7] ^ rk[14];
+			rk += 8;
+		}
+		break;
+
+	default:
+		psTraceCrypto("Invalid AES key length\n");
+		return CRYPT_INVALID_KEYSIZE;
+	}
+
+	switch(flags) {
+
+	case PS_AES_ENCRYPT:
+		key->type = PS_AES_ENCRYPT;
+		break;
+
+	case PS_AES_DECRYPT:
+		key->type = PS_AES_DECRYPT;
+		/* Replace the key schedule with the inverse key schedule */
+		invertKeySchedule(key);
+		break;
+
+	default:
+		return PS_ARG_FAIL;
+	}
+
+	return PS_SUCCESS;
+}
+
+/******************************************************************************/
+
+void psAesClearBlockKey(psAesKey_t *key)
+{
+    memset_s(key, sizeof(psAesKey_t), 0x0, sizeof(psAesKey_t));
+}
+
+/******************************************************************************/
+/*
+	AES constants
+*/
+
 /* The precomputed tables for AES */
 /*
 Te0[x] = S [x].[02, 01, 01, 03];
@@ -120,6 +284,7 @@ static const uint32 TE0[] = {
 	0x7bb0b0cbUL, 0xa85454fcUL, 0x6dbbbbd6UL, 0x2c16163aUL,
 };
 
+#ifndef PS_AES_IMPROVE_PERF_INCREASE_CODESIZE
 static const uint32 Te4[] = {
 	0x63636363UL, 0x7c7c7c7cUL, 0x77777777UL, 0x7b7b7b7bUL,
 	0xf2f2f2f2UL, 0x6b6b6b6bUL, 0x6f6f6f6fUL, 0xc5c5c5c5UL,
@@ -186,6 +351,7 @@ static const uint32 Te4[] = {
 	0x41414141UL, 0x99999999UL, 0x2d2d2d2dUL, 0x0f0f0f0fUL,
 	0xb0b0b0b0UL, 0x54545454UL, 0xbbbbbbbbUL, 0x16161616UL,
 };
+#endif /* PS_AES_IMPROVE_PERF_INCREASE_CODESIZE */
 
 static const uint32 TD0[] = {
 	0x51f4a750UL, 0x7e416553UL, 0x1a17a4c3UL, 0x3a275e96UL,
@@ -322,34 +488,7 @@ static const uint32 Td4[] = {
 	0xe1f27f3aUL, 0xf5710fb0UL, 0xada0e5c4UL, 0x98e4c919UL
 };
 
-#ifndef PS_AES_IMPROVE_PERF_INCREASE_CODESIZE
-
-#define Te0(x) TE0[x]
-#define Te1(x) ROR(TE0[x], 8)
-#define Te2(x) ROR(TE0[x], 16)
-#define Te3(x) ROR(TE0[x], 24)
-
-#define Td0(x) TD0[x]
-#define Td1(x) ROR(TD0[x], 8)
-#define Td2(x) ROR(TD0[x], 16)
-#define Td3(x) ROR(TD0[x], 24)
-
-#define Te4_0 0x000000FF & Te4
-#define Te4_1 0x0000FF00 & Te4
-#define Te4_2 0x00FF0000 & Te4
-#define Te4_3 0xFF000000 & Te4
-
-#else /* PS_AES_IMPROVE_PERF_INCREASE_CODESIZE */
-
-#define Te0(x) TE0[x]
-#define Te1(x) TE1[x]
-#define Te2(x) TE2[x]
-#define Te3(x) TE3[x]
-
-#define Td0(x) TD0[x]
-#define Td1(x) TD1[x]
-#define Td2(x) TD2[x]
-#define Td3(x) TD3[x]
+#ifdef PS_AES_IMPROVE_PERF_INCREASE_CODESIZE
 
 static const uint32 TE1[256] = {
 	0xa5c66363UL, 0x84f87c7cUL, 0x99ee7777UL, 0x8df67b7bUL,
@@ -1032,369 +1171,63 @@ static const uint32 Tks3[] = {
 
 #endif /* PS_AES_IMPROVE_PERF_INCREASE_CODESIZE */
 
-static const uint32 rcon[] = {
-	0x01000000UL, 0x02000000UL, 0x04000000UL, 0x08000000UL,
-	0x10000000UL, 0x20000000UL, 0x40000000UL, 0x80000000UL,
-	0x1B000000UL, 0x36000000UL, /* for 128-bit blocks, AES never uses more than 10 rcon values */
-};
+/******************************************************************************/
+/* Macros for above */
 
-#ifdef USE_BURN_STACK
-static void _aes_ecb_decrypt(const unsigned char *ct, unsigned char *pt,
-				psAesKey_t *skey);
-static void _aes_ecb_encrypt(const unsigned char *pt, unsigned char *ct,
-				psAesKey_t *skey);
+#ifndef PS_AES_IMPROVE_PERF_INCREASE_CODESIZE
+#define Te0(x) TE0[x]
+#define Te1(x) ROR(TE0[x], 8)
+#define Te2(x) ROR(TE0[x], 16)
+#define Te3(x) ROR(TE0[x], 24)
+#define Td0(x) TD0[x]
+#define Td1(x) ROR(TD0[x], 8)
+#define Td2(x) ROR(TD0[x], 16)
+#define Td3(x) ROR(TD0[x], 24)
+#define Te4_0 0x000000FF & Te4
+#define Te4_1 0x0000FF00 & Te4
+#define Te4_2 0x00FF0000 & Te4
+#define Te4_3 0xFF000000 & Te4
+#else
+#define Te0(x) TE0[x]
+#define Te1(x) TE1[x]
+#define Te2(x) TE2[x]
+#define Te3(x) TE3[x]
+#define Td0(x) TD0[x]
+#define Td1(x) TD1[x]
+#define Td2(x) TD2[x]
+#define Td3(x) TD3[x]
 #endif
 
-static uint32 setup_mix(uint32 temp)
-{
-	return	(Te4_3[byte(temp, 2)]) ^
-			(Te4_2[byte(temp, 1)]) ^
-			(Te4_1[byte(temp, 0)]) ^
-			(Te4_0[byte(temp, 3)]);
-}
-
-#ifndef PS_AES_IMPROVE_PERF_INCREASE_CODESIZE
-static uint32 setup_mix2(uint32 temp)
-{
-	return	Td0(255 & Te4[byte(temp, 3)]) ^
-			Td1(255 & Te4[byte(temp, 2)]) ^
-			Td2(255 & Te4[byte(temp, 1)]) ^
-			Td3(255 & Te4[byte(temp, 0)]);
-}
-#endif /* PS_AES_IMPROVE_PERF_INCREASE_CODESIZE */
-
-/*
-	Software implementation of AES CBC APIs
- */
-#ifndef USE_AES_CBC_EXTERNAL
-int32 psAesInit(psCipherContext_t *ctx, unsigned char *IV,
-				  unsigned char *key, uint32 keylen)
-{
-	int32		x, err;
-
-	if (IV == NULL || key == NULL || ctx == NULL) {
-		psTraceCrypto("psAesInit arg fail\n");
-		return PS_ARG_FAIL;
-	}
-	memset(ctx, 0x0, sizeof(psCipherContext_t));
-/*
-	setup cipher
- */
-	if ((err = psAesInitKey(key, keylen, &ctx->aes.key)) != PS_SUCCESS) {
-		return err;
-	}
-/*
-	copy IV
- */
-	ctx->aes.blocklen = 16;
-	for (x = 0; x < ctx->aes.blocklen; x++) {
-		ctx->aes.IV[x] = IV[x];
-	}
-	return PS_SUCCESS;
-}
-
-int32 psAesEncrypt(psCipherContext_t *ctx, unsigned char *pt,
-					 unsigned char *ct, uint32 len)
-{
-	int32			x;
-	uint32			i;
-	unsigned char	tmp[MAXBLOCKSIZE];
-
-	if (pt == NULL || ct == NULL || ctx == NULL || (len & 0x7) != 0) {
-		psTraceCrypto("Bad parameters to psAesEncrypt\n");
-		return PS_ARG_FAIL;
-	}
-
-/*
-	is blocklen valid?
- */
-	if (ctx->aes.blocklen < 0 || (ctx->aes.blocklen >
-			(int32)sizeof(ctx->aes.IV))) {
-		psTraceCrypto("Bad blocklen in psAesEncrypt\n");
-		return PS_LIMIT_FAIL;
-	}
-
-	for (i = 0; i < len; i += ctx->aes.blocklen) {
-/*
-		xor IV against plaintext
- */
-		for (x = 0; x < ctx->aes.blocklen; x++) {
-			tmp[x] = pt[x] ^ ctx->aes.IV[x];
-		}
-/*
-		encrypt
- */
-		psAesEncryptBlock(tmp, ct, &ctx->aes.key);
-
-/*
-		store IV [ciphertext] for a future block
- */
-		for (x = 0; x < ctx->aes.blocklen; x++) {
-			ctx->aes.IV[x] = ct[x];
-		}
-		ct += ctx->aes.blocklen;
-		pt += ctx->aes.blocklen;
-	}
-
-	memset_s(tmp, sizeof(tmp), 0x0, sizeof(tmp));
-	return len;
-}
-
-int32 psAesDecrypt(psCipherContext_t *ctx, unsigned char *ct,
-					 unsigned char *pt, uint32 len)
-{
-	int32			x;
-	uint32			i;
-	unsigned char	tmp[MAXBLOCKSIZE], tmp2[MAXBLOCKSIZE];
-
-	if (pt == NULL || ct == NULL || ctx == NULL || (len & 0x7) != 0) {
-		psTraceCrypto("Bad parameters to psAesDecrypt\n");
-		return PS_ARG_FAIL;
-	}
-
-/*
-	is blocklen valid?
- */
-	if (ctx->aes.blocklen < 0 || (ctx->aes.blocklen >
-			(int32)sizeof(ctx->aes.IV))) {
-		psTraceCrypto("Bad blocklen in psAesDecrypt\n");
-		return PS_LIMIT_FAIL;
-	}
-	for (i = 0; i < len; i += ctx->aes.blocklen) {
-/*
-		decrypt the block from ct into tmp
- */
-		psAesDecryptBlock(ct, tmp, &ctx->aes.key);
-/*
-		xor IV against the plaintext of the previous step
- */
-		for (x = 0; x < ctx->aes.blocklen; x++) {
-/*
-			copy CT in case ct == pt
- */
-			tmp2[x] = ct[x];
-/*
-			actually decrypt the byte
- */
-			pt[x] =	tmp[x] ^ ctx->aes.IV[x];
-		}
-/*
-		replace IV with this current ciphertext
- */
-		for (x = 0; x < ctx->aes.blocklen; x++) {
-			ctx->aes.IV[x] = tmp2[x];
-		}
-		ct += ctx->aes.blocklen;
-		pt += ctx->aes.blocklen;
-	}
-	memset_s(tmp, sizeof(tmp), 0x0, sizeof(tmp));
-	memset_s(tmp2, sizeof(tmp2), 0x0, sizeof(tmp2));
-
-	return len;
-}
-
-#endif /* USE_AES_CBC_EXTERNAL */
-
 /******************************************************************************/
-/*
-	Initialize the AES (Rijndael) block cipher
-
-	key: The symmetric key you wish to pass
-	keylen:  The key length in bytes
-	skey: The key in as scheduled by this function.
-*/
-
-int32 psAesInitKey(const unsigned char *key, uint32 keylen, psAesKey_t *skey)
-{
-	int32		i, j;
-	uint32		temp, *rk, *rrk;
-
-	if (key == NULL || skey == NULL) {
-		psTraceCrypto("Bad args to psAesInitKey\n");
-		return PS_ARG_FAIL;
-	}
-
-	if (keylen != 16 && keylen != 24 && keylen != 32) {
-		psTraceCrypto("Invalid AES key length\n");
-		return CRYPT_INVALID_KEYSIZE;
-	}
-
-	memset(skey, 0x0, sizeof(psAesKey_t));
-	skey->Nr = 10 + ((keylen/8)-2)*2;
-
-/*
-	setup the forward key
- */
-	i				= 0;
-	rk				= skey->eK;
-	LOAD32H(rk[0], key	   );
-	LOAD32H(rk[1], key +  4);
-	LOAD32H(rk[2], key +  8);
-	LOAD32H(rk[3], key + 12);
-	if (keylen == 16) {
-		j = 44;
-		for (;;) {
-			temp  = rk[3];
-			rk[4] = rk[0] ^ setup_mix(temp) ^ rcon[i];
-			rk[5] = rk[1] ^ rk[4];
-			rk[6] = rk[2] ^ rk[5];
-			rk[7] = rk[3] ^ rk[6];
-			if (++i == 10) {
-				break;
-			}
-			rk += 4;
-		}
-	} else if (keylen == 24) {
-		j = 52;
-		LOAD32H(rk[4], key + 16);
-		LOAD32H(rk[5], key + 20);
-		for (;;) {
-		#ifdef _MSC_VER
-			temp = skey->eK[rk - skey->eK + 5];
-		#else
-			temp = rk[5];
-		#endif /* _MSC_VER */
-			rk[ 6] = rk[ 0] ^ setup_mix(temp) ^ rcon[i];
-			rk[ 7] = rk[ 1] ^ rk[ 6];
-			rk[ 8] = rk[ 2] ^ rk[ 7];
-			rk[ 9] = rk[ 3] ^ rk[ 8];
-			if (++i == 8) {
-				break;
-			}
-			rk[10] = rk[ 4] ^ rk[ 9];
-			rk[11] = rk[ 5] ^ rk[10];
-			rk += 6;
-		}
-	} else if (keylen == 32) {
-		j = 60;
-		LOAD32H(rk[4], key + 16);
-		LOAD32H(rk[5], key + 20);
-		LOAD32H(rk[6], key + 24);
-		LOAD32H(rk[7], key + 28);
-		for (;;) {
-		#ifdef _MSC_VER
-			temp = skey->eK[rk - skey->eK + 7];
-		#else
-			temp = rk[7];
-		#endif /* _MSC_VER */
-			rk[ 8] = rk[ 0] ^ setup_mix(temp) ^ rcon[i];
-			rk[ 9] = rk[ 1] ^ rk[ 8];
-			rk[10] = rk[ 2] ^ rk[ 9];
-			rk[11] = rk[ 3] ^ rk[10];
-			if (++i == 7) {
-				break;
-			}
-			temp = rk[11];
-			rk[12] = rk[ 4] ^ setup_mix(ROR(temp, 8));
-			rk[13] = rk[ 5] ^ rk[12];
-			rk[14] = rk[ 6] ^ rk[13];
-			rk[15] = rk[ 7] ^ rk[14];
-			rk += 8;
-		}
-	} else {
-/*
-		this can't happen
- */
-		return PS_FAILURE;
-	}
-
-/*
-	setup the inverse key now
- */
-	rk		= skey->dK;
-	rrk		= skey->eK + j - 4;
-
-/*
-	apply the inverse MixColumn transform to all round keys but
-	the first and the last:
- */
-	/* copy first */
-	*rk++	= *rrk++;
-	*rk++	= *rrk++;
-	*rk++	= *rrk++;
-	*rk		= *rrk;
-	rk -= 3; rrk -= 3;
-
-	for (i = 1; i < skey->Nr; i++) {
-		rrk	-= 4;
-		rk	+= 4;
-#ifndef PS_AES_IMPROVE_PERF_INCREASE_CODESIZE
-		temp = rrk[0];
-		rk[0] = setup_mix2(temp);
-		temp = rrk[1];
-		rk[1] = setup_mix2(temp);
-		temp = rrk[2];
-		rk[2] = setup_mix2(temp);
-		temp = rrk[3];
-		rk[3] = setup_mix2(temp);
-#else /* PS_AES_IMPROVE_PERF_INCREASE_CODESIZE */
-		temp = rrk[0];
-		rk[0] =
-			Tks0[byte(temp, 3)] ^
-			Tks1[byte(temp, 2)] ^
-			Tks2[byte(temp, 1)] ^
-			Tks3[byte(temp, 0)];
-		temp = rrk[1];
-		rk[1] =
-			Tks0[byte(temp, 3)] ^
-			Tks1[byte(temp, 2)] ^
-			Tks2[byte(temp, 1)] ^
-			Tks3[byte(temp, 0)];
-		temp = rrk[2];
-		rk[2] =
-			Tks0[byte(temp, 3)] ^
-			Tks1[byte(temp, 2)] ^
-			Tks2[byte(temp, 1)] ^
-			Tks3[byte(temp, 0)];
-		temp = rrk[3];
-		rk[3] =
-			Tks0[byte(temp, 3)] ^
-			Tks1[byte(temp, 2)] ^
-			Tks2[byte(temp, 1)] ^
-			Tks3[byte(temp, 0)];
-#endif /* PS_AES_IMPROVE_PERF_INCREASE_CODESIZE */
-	}
-
-	/* copy last */
-	rrk	-= 4;
-	rk	+= 4;
-	*rk++ = *rrk++;
-	*rk++ = *rrk++;
-	*rk++ = *rrk++;
-	*rk	  = *rrk;
-
-	return PS_SUCCESS;
-}
-
-
 #ifdef USE_BURN_STACK
-void psAesEncryptBlock(const unsigned char *pt, unsigned char *ct,
-				psAesKey_t *skey)
+static void _aes_ecb_encrypt(psAesKey_t *key, const unsigned char *pt,
+				unsigned char *ct);
+void psAesEncryptBlock(psAesKey_t *key, const unsigned char *pt,
+						unsigned char *ct)
 {
-	_aes_ecb_encrypt(pt, ct, skey);
+	_aes_ecb_encrypt(key, pt, ct);
 	psBurnStack(sizeof(uint32)*8 + sizeof(uint32*) + sizeof(int32)*2);
 }
-static void _aes_ecb_encrypt(const unsigned char *pt, unsigned char *ct,
-				psAesKey_t *skey)
+static void _aes_ecb_encrypt(psAesKey_t *key, const unsigned char *pt,
+								unsigned char *ct)
 #else
-void psAesEncryptBlock(const unsigned char *pt, unsigned char *ct,
-				psAesKey_t *skey)
+void psAesEncryptBlock(psAesKey_t *key, const unsigned char *pt,
+						unsigned char *ct)
 #endif /* USE_BURN_STACK */
 {
-	uint32	s0, s1, s2, s3, t0, t1, t2, t3, *rk;
-	int32	Nr, r;
+	uint32_t	s0, s1, s2, s3, t0, t1, t2, t3, *rk;
+	int			Nr, r;
 
-	if (pt == NULL || ct == NULL || skey == NULL) {
+#ifdef CRYPTO_ASSERT
+	if (pt == NULL || ct == NULL || key == NULL || key->type != PS_AES_ENCRYPT) {
 		return;
 	}
+#endif
 
-	Nr = skey->Nr;
-	rk = skey->eK;
+	Nr = key->rounds;
+	rk = key->skey;
 
-/*
-	map byte array block to cipher state
-	and add initial round key:
- */
+	/* Map byte array block to cipher state and add initial round key */
 	LOAD32H(s0, pt     ); s0 ^= rk[0];
 	LOAD32H(s1, pt +  4); s1 ^= rk[1];
 	LOAD32H(s2, pt +  8); s2 ^= rk[2];
@@ -1427,7 +1260,7 @@ void psAesEncryptBlock(const unsigned char *pt, unsigned char *ct,
 			Te2(byte(s1, 1)) ^
 			Te3(byte(s2, 0)) ^
 			rk[3];
-		if (r == Nr-2) {
+		if (r == Nr - 2) {
 			break;
 		}
 		s0 = t0; s1 = t1; s2 = t2; s3 = t3;
@@ -1435,9 +1268,7 @@ void psAesEncryptBlock(const unsigned char *pt, unsigned char *ct,
 	rk += 4;
 #else /* PS_AES_IMPROVE_PERF_INCREASE_CODESIZE */
 
-/*
-	Nr - 1 full rounds:
- */
+	/* Nr - 1 full rounds */
 	r = Nr >> 1;
 	for (;;) {
 
@@ -1498,9 +1329,7 @@ void psAesEncryptBlock(const unsigned char *pt, unsigned char *ct,
 	}
 #endif /* PS_AES_IMPROVE_PERF_INCREASE_CODESIZE */
 
-/*
-		apply last round and map cipher state to byte array block:
- */
+	/* Apply last round and map cipher state to byte array block */
 	s0 =
 		(Te4_3[byte(t0, 3)]) ^
 		(Te4_2[byte(t1, 2)]) ^
@@ -1532,32 +1361,34 @@ void psAesEncryptBlock(const unsigned char *pt, unsigned char *ct,
 }
 
 #ifdef USE_BURN_STACK
-void psAesDecryptBlock(const unsigned char *ct, unsigned char *pt,
-				psAesKey_t *skey)
+static void _aes_ecb_decrypt(psAesKey_t *key, const unsigned char *ct,
+				unsigned char *pt);
+void psAesDecryptBlock(psAesKey_t *key, const unsigned char *ct,
+				unsigned char *pt)
 {
-	_aes_ecb_decrypt(ct, pt, skey);
+	_aes_ecb_decrypt(key, ct, pt);
 	psBurnStack(sizeof(uint32)*8 + sizeof(uint32*) + sizeof(int32)*2);
 }
-static void _aes_ecb_decrypt(const unsigned char *ct, unsigned char *pt,
-				psAesKey_t *skey)
+static void _aes_ecb_decrypt(psAesKey_t *key, const unsigned char *ct,
+				unsigned char *pt)
 #else
-void psAesDecryptBlock(const unsigned char *ct, unsigned char *pt,
-				psAesKey_t *skey)
+void psAesDecryptBlock(psAesKey_t *key, const unsigned char *ct,
+				unsigned char *pt)
 #endif /* USE_BURN_STACK */
 {
 	uint32		s0, s1, s2, s3, t0, t1, t2, t3, *rk;
 	int32		Nr, r;
 
-	if (pt == NULL || ct == NULL || skey == NULL) {
+#ifdef CRYPTO_ASSERT
+	if (pt == NULL || ct == NULL || key == NULL || key->type != PS_AES_DECRYPT) {
 		return;
 	}
+#endif
 
-	Nr = skey->Nr;
-	rk = skey->dK;
+	Nr = key->rounds;
+	rk = key->skey;
 
-/*
-	map byte array block to cipher state and add initial round key:
- */
+	/* Map byte array block to cipher state and add initial round key */
 	LOAD32H(s0, ct		); s0 ^= rk[0];
 	LOAD32H(s1, ct	+  4); s1 ^= rk[1];
 	LOAD32H(s2, ct	+  8); s2 ^= rk[2];
@@ -1590,7 +1421,7 @@ void psAesDecryptBlock(const unsigned char *ct, unsigned char *pt,
 			Td2(byte(s1, 1)) ^
 			Td3(byte(s0, 0)) ^
 			rk[3];
-		if (r == Nr-2) {
+		if (r == Nr - 2) {
 			break;
 		}
 		s0 = t0; s1 = t1; s2 = t2; s3 = t3;
@@ -1599,9 +1430,7 @@ void psAesDecryptBlock(const unsigned char *ct, unsigned char *pt,
 
 #else /* PS_AES_IMPROVE_PERF_INCREASE_CODESIZE */
 
-/*
-	Nr - 1 full rounds:
- */
+	/* Nr - 1 full rounds */
 	r = Nr >> 1;
 	for	(;;) {
 
@@ -1662,9 +1491,7 @@ void psAesDecryptBlock(const unsigned char *ct, unsigned char *pt,
 	}
 #endif /* PS_AES_IMPROVE_PERF_INCREASE_CODESIZE */
 
-/*
-	apply last round and map cipher state to byte array block:
- */
+	/* Apply last round and map cipher state to byte array block */
 	s0 =
 		(Td4[byte(t0, 3)] & 0xff000000) ^
 		(Td4[byte(t3, 2)] & 0x00ff0000) ^
@@ -1695,10 +1522,119 @@ void psAesDecryptBlock(const unsigned char *ct, unsigned char *pt,
 	STORE32H(s3, pt+12);
 }
 
+/******************************************************************************/
+
+static uint32 setup_mix(uint32 temp)
+{
+	return	(Te4_3[byte(temp, 2)]) ^
+			(Te4_2[byte(temp, 1)]) ^
+			(Te4_1[byte(temp, 0)]) ^
+			(Te4_0[byte(temp, 3)]);
+}
+
+#ifndef PS_AES_IMPROVE_PERF_INCREASE_CODESIZE
+static uint32 setup_mix2(uint32 temp)
+{
+	return	Td0(255 & Te4[byte(temp, 3)]) ^
+			Td1(255 & Te4[byte(temp, 2)]) ^
+			Td2(255 & Te4[byte(temp, 1)]) ^
+			Td3(255 & Te4[byte(temp, 0)]);
+}
+#endif /* PS_AES_IMPROVE_PERF_INCREASE_CODESIZE */
 
 /******************************************************************************/
-/******************************************************************************/
-#endif /* !USE_AES_CBC_EXTERNAL */
-#endif /* USE_AES */
+/**
+	Convert a forward (encrypt) key schedule into an inverse (decrypt)
+	key schedule.
+	@note Inline since it is only called from one internal function, and
+	it can save some stack memory. Compiler optimization would probably
+	inline this on its own anyway.
+*/
+__inline static void invertKeySchedule(psAesKey_t *key)
+{
+	int			i;
+	uint32		temp, *rk, *rrk;
+	uint32		ek[64];
+
+	/* Copy the current (encrypt key) to a temp buffer */
+	memcpy(ek, key->skey, sizeof(ek));
+	rk = key->skey;
+	switch (key->rounds) {
+	case 10:
+		rrk = ek + 40;
+		break;
+	case 12:
+		rrk = ek + 48;
+		break;
+	case 14:
+		rrk = ek + 56;
+		break;
+	default:
+		return;
+	}
+/*
+	apply the inverse MixColumn transform to all round keys but
+	the first and the last:
+ */
+	/* copy first */
+	*rk++	= *rrk++;
+	*rk++	= *rrk++;
+	*rk++	= *rrk++;
+	*rk		= *rrk;
+	rk -= 3; rrk -= 3;
+
+	for (i = 1; i < key->rounds; i++) {
+		rrk	-= 4;
+		rk	+= 4;
+#ifndef PS_AES_IMPROVE_PERF_INCREASE_CODESIZE
+		temp = rrk[0];
+		rk[0] = setup_mix2(temp);
+		temp = rrk[1];
+		rk[1] = setup_mix2(temp);
+		temp = rrk[2];
+		rk[2] = setup_mix2(temp);
+		temp = rrk[3];
+		rk[3] = setup_mix2(temp);
+#else /* PS_AES_IMPROVE_PERF_INCREASE_CODESIZE */
+		temp = rrk[0];
+		rk[0] =
+			Tks0[byte(temp, 3)] ^
+			Tks1[byte(temp, 2)] ^
+			Tks2[byte(temp, 1)] ^
+			Tks3[byte(temp, 0)];
+		temp = rrk[1];
+		rk[1] =
+			Tks0[byte(temp, 3)] ^
+			Tks1[byte(temp, 2)] ^
+			Tks2[byte(temp, 1)] ^
+			Tks3[byte(temp, 0)];
+		temp = rrk[2];
+		rk[2] =
+			Tks0[byte(temp, 3)] ^
+			Tks1[byte(temp, 2)] ^
+			Tks2[byte(temp, 1)] ^
+			Tks3[byte(temp, 0)];
+		temp = rrk[3];
+		rk[3] =
+			Tks0[byte(temp, 3)] ^
+			Tks1[byte(temp, 2)] ^
+			Tks2[byte(temp, 1)] ^
+			Tks3[byte(temp, 0)];
+#endif /* PS_AES_IMPROVE_PERF_INCREASE_CODESIZE */
+	}
+
+	/* copy last */
+	rrk	-= 4;
+	rk	+= 4;
+	*rk++ = *rrk++;
+	*rk++ = *rrk++;
+	*rk++ = *rrk++;
+	*rk	  = *rrk;
+
+	memset_s(ek, sizeof(ek), 0x0, sizeof(ek));
+}
+
+#endif /* USE_MATRIX_AES_BLOCK */
+
 /******************************************************************************/
 
