@@ -1,11 +1,11 @@
 /**
  *	@file    seed.c
- *	@version 33ef80f (HEAD, tag: MATRIXSSL-3-7-2-OPEN, tag: MATRIXSSL-3-7-2-COMM, origin/master, origin/HEAD, master)
+ *	@version 7f5c362 (HEAD -> master)
  *
  *	seed implementation of SEED derived from RFC4269.
  */
 /*
- *	Copyright (c) 2013-2015 INSIDE Secure Corporation
+ *	Copyright (c) 2013-2016 INSIDE Secure Corporation
  *	Copyright (c) PeerSec Networks, 2002-2011
  *	All Rights Reserved
  *
@@ -19,7 +19,7 @@
  *	This General Public License does NOT permit incorporating this software
  *	into proprietary programs.  If you are unable to comply with the GPL, a
  *	commercial license for this software may be purchased from INSIDE at
- *	http://www.insidesecure.com/eng/Company/Locations
+ *	http://www.insidesecure.com/
  *
  *	This program is distributed in WITHOUT ANY WARRANTY; without even the
  *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
@@ -34,11 +34,18 @@
 
 #include "../cryptoApi.h"
 
-/******************************************************************************/
-#ifdef USE_SEED
+#ifdef USE_MATRIX_SEED
+
 /******************************************************************************/
 
-static const uint32 SS0[256] = {
+static int32_t psSeedInitKey(const unsigned char key[SEED_KEYLEN],
+				psSeedKey_t *skey);
+static void psSeedEncryptBlock(const unsigned char *pt, unsigned char *ct,
+				psSeedKey_t *skey);
+static void psSeedDecryptBlock(const unsigned char *ct, unsigned char *pt,
+				psSeedKey_t *skey);
+
+static const uint32_t SS0[256] = {
 0x2989A1A8UL,0x05858184UL,0x16C6D2D4UL,0x13C3D3D0UL,0x14445054UL,0x1D0D111CUL,0x2C8CA0ACUL,0x25052124UL,
 0x1D4D515CUL,0x03434340UL,0x18081018UL,0x1E0E121CUL,0x11415150UL,0x3CCCF0FCUL,0x0ACAC2C8UL,0x23436360UL,
 0x28082028UL,0x04444044UL,0x20002020UL,0x1D8D919CUL,0x20C0E0E0UL,0x22C2E2E0UL,0x08C8C0C8UL,0x17071314UL,
@@ -197,156 +204,118 @@ static const uint32 KCi[16] = {
    L2 ^= T; \
    L1 ^= (T + G(T2 + (R1 ^ K1))); \
 
-
 /******************************************************************************/
 /*
 	Init the SEED block cipher context for CBC-EDE mode.
 	IV should point to 16 bytes of initialization vector
 	Key should point to 16 bytes of data
 */
-int32 psSeedInit(psCipherContext_t *ctx, unsigned char *IV,
-				   unsigned char *key, uint32 keylen)
+int32_t psSeedInit(psSeed_t *seed, const unsigned char IV[SEED_IVLEN],
+				   const unsigned char key[SEED_KEYLEN])
 {
-	int32 x, err;
+	int32_t		err;
+	uint8_t		x;
 
-	if (IV == NULL || key == NULL || ctx == NULL || keylen != SSL_SEED_KEY_LEN){
+#ifdef CRYPTO_ASSERT
+	if (IV == NULL || key == NULL || seed == NULL){
 		psTraceCrypto("Bad parameter failure in psSeedInit\n");
 		return PS_ARG_FAIL;
 	}
-/*
-	setup cipher
- */
-	if ((err = psSeedInitKey(key, keylen, &ctx->seed.key)) != PS_SUCCESS) {
+#endif
+	/* setup cipher */
+	if ((err = psSeedInitKey(key, &seed->key)) != PS_SUCCESS) {
 		return err;
 	}
-/*
-	copy IV
- */
-	ctx->seed.blocklen = SSL_SEED_IV_LEN;
-	for (x = 0; x < ctx->seed.blocklen; x++) {
-		ctx->seed.IV[x] = IV[x];
+	/* copy IV */
+	for (x = 0; x < SEED_BLOCKLEN; x++) {
+		seed->IV[x] = IV[x];
 	}
 	return PS_SUCCESS;
 }
 
+/******************************************************************************/
 
-int32 psSeedEncrypt(psCipherContext_t *ctx, unsigned char *pt,
-					 unsigned char *ct, uint32 len)
+void psSeedEncrypt(psSeed_t *seed, const unsigned char *pt,
+					 unsigned char *ct, uint32_t len)
 {
-	int32			x;
-	uint32			i;
-	unsigned char	tmp[MAXBLOCKSIZE];
+	uint8_t			i, x;
+	unsigned char	tmp[SEED_BLOCKLEN];
 
-	if (pt == NULL || ct == NULL || ctx == NULL || (len & 0x7) != 0) {
+#ifdef CRYPTO_ASSERT
+	if (pt == NULL || ct == NULL || seed == NULL || (len & 0x7) != 0) {
 		psTraceCrypto("Bad parameters to psSeedEncrypt\n");
-		return PS_ARG_FAIL;
+		return;
 	}
-
-/*
-	is blocklen valid?
- */
-	if (ctx->seed.blocklen < 0 || (ctx->seed.blocklen >
-			(int32)sizeof(ctx->seed.IV))) {
-		psTraceCrypto("Blocklen test failure in psSeedEncrypt\n");
-		return PS_LIMIT_FAIL;
-	}
-
-	for (i = 0; i < len; i += ctx->seed.blocklen) {
-/*
-		xor IV against plaintext
- */
-		for (x = 0; x < ctx->seed.blocklen; x++) {
-			tmp[x] = pt[x] ^ ctx->seed.IV[x];
+#endif
+	for (i = 0; i < len; i += SEED_BLOCKLEN) {
+		for (x = 0; x < SEED_BLOCKLEN; x++) {
+			tmp[x] = pt[x] ^ seed->IV[x];
 		}
-/*
-		encrypt
- */
-		psSeedEncryptBlock(tmp, ct, &ctx->seed.key);
-
-/*
-		store IV [ciphertext] for a future block
- */
-		for (x = 0; x < ctx->seed.blocklen; x++) {
-			ctx->seed.IV[x] = ct[x];
+		psSeedEncryptBlock(tmp, ct, &seed->key);
+		/* store IV [ciphertext] for a future block */
+		for (x = 0; x < SEED_BLOCKLEN; x++) {
+			seed->IV[x] = ct[x];
 		}
-		ct += ctx->seed.blocklen;
-		pt += ctx->seed.blocklen;
+		ct += SEED_BLOCKLEN;
+		pt += SEED_BLOCKLEN;
 	}
-
-	memset(tmp, 0x0, sizeof(tmp));
-	return len;
+	memzero_s(tmp, sizeof(tmp));
 }
 
-int32 psSeedDecrypt(psCipherContext_t *ctx, unsigned char *ct,
-					 unsigned char *pt, uint32 len)
+/******************************************************************************/
+
+void psSeedDecrypt(psSeed_t *seed, const unsigned char *ct,
+					 unsigned char *pt, uint32_t len)
 {
-	int32			x;
-	uint32			i;
-	unsigned char	tmp[MAXBLOCKSIZE], tmp2[MAXBLOCKSIZE];
+	uint8_t			i, x;
+	unsigned char	tmp[SEED_BLOCKLEN], tmp2[SEED_BLOCKLEN];
 
-	if (pt == NULL || ct == NULL || ctx == NULL || (len & 0x7) != 0) {
+#ifdef CRYPTO_ASSERT
+	if (pt == NULL || ct == NULL || seed == NULL || (len & 0x7) != 0) {
 		psTraceCrypto("Bad parameters to psSeedDecrypt\n");
-		return PS_ARG_FAIL;
+		return;
 	}
-
-/*
-	is blocklen valid?
- */
-	if (ctx->seed.blocklen < 0 || (ctx->seed.blocklen >
-			(int32)sizeof(ctx->seed.IV))) {
-		psTraceCrypto("Blocklen test failure in psSeedDecrypt\n");
-		return PS_LIMIT_FAIL;
-	}
-	for (i = 0; i < len; i += ctx->seed.blocklen) {
-/*
-		decrypt the block from ct into tmp
- */
-		psSeedDecryptBlock(ct, tmp, &ctx->seed.key);
-/*
-		xor IV against the plaintext of the previous step
- */
-		for (x = 0; x < ctx->seed.blocklen; x++) {
-/*
-			copy CT in case ct == pt
- */
+#endif
+	for (i = 0; i < len; i += SEED_BLOCKLEN) {
+		/* decrypt the block from ct into tmp */
+		psSeedDecryptBlock(ct, tmp, &seed->key);
+		/* xor IV against the plaintext of the previous step */
+		for (x = 0; x < SEED_BLOCKLEN; x++) {
+			/* copy CT in case ct == pt */
 			tmp2[x] = ct[x];
-/*
-			actually decrypt the byte
- */
-			pt[x] =	tmp[x] ^ ctx->seed.IV[x];
+			/* actually decrypt the byte */
+			pt[x] =	tmp[x] ^ seed->IV[x];
 		}
-/*
-		replace IV with this current ciphertext
- */
-		for (x = 0; x < ctx->seed.blocklen; x++) {
-			ctx->seed.IV[x] = tmp2[x];
+		/* replace IV with this current ciphertext */
+		for (x = 0; x < SEED_BLOCKLEN; x++) {
+			seed->IV[x] = tmp2[x];
 		}
-		ct += ctx->seed.blocklen;
-		pt += ctx->seed.blocklen;
+		ct += SEED_BLOCKLEN;
+		pt += SEED_BLOCKLEN;
 	}
-	memset(tmp, 0x0, sizeof(tmp));
-	memset(tmp2, 0x0, sizeof(tmp2));
-	return len;
+	memzero_s(tmp, sizeof(tmp));
+	memzero_s(tmp2, sizeof(tmp2));
 }
 
+/******************************************************************************/
 
- /**
+void psSeedClear(psSeed_t *seed)
+{
+	memzero_s(seed, sizeof(psSeed_t));
+}
+
+/******************************************************************************/
+/**
 	Initialize the SEED block cipher
 	@param key The symmetric key you wish to pass
 	@param keylen The key length in bytes
 	@param skey The key in as scheduled by this function.
  */
-int32 psSeedInitKey(const unsigned char *key, uint32 keylen, psSeedKey_t *skey)
+int32_t psSeedInitKey(const unsigned char *key, psSeedKey_t *skey)
 {
 	int32	i;
 	uint32	tmp, k1, k2, k3, k4;
 
-	if (keylen != 16) {
-		psTraceIntCrypto("Invalid key size error in psSeedInitKey: %d\n",
-			(int32)keylen);
-		return CRYPT_INVALID_KEYSIZE;
-	}
-
 	/* load key */
 	LOAD32H(k1, key);
 	LOAD32H(k2, key+4);
@@ -384,14 +353,15 @@ static void rounds(uint32 *P, uint32 *K)
 	}
 }
 
+/******************************************************************************/
 /**
   Encrypts a block of text with SEED
   @param pt The input plaintext (16 bytes)
   @param ct The output ciphertext (16 bytes)
   @param skey The key as scheduled
 */
-void psSeedEncryptBlock(const unsigned char *pt, unsigned char *ct,
-						psSeedKey_t *skey)
+static void psSeedEncryptBlock(const unsigned char *pt, unsigned char *ct,
+								psSeedKey_t *skey)
 {
 	uint32 P[4];
 	LOAD32H(P[0], pt);
@@ -403,17 +373,17 @@ void psSeedEncryptBlock(const unsigned char *pt, unsigned char *ct,
 	STORE32H(P[3], ct+4);
 	STORE32H(P[0], ct+8);
 	STORE32H(P[1], ct+12);
-	return;
 }
 
+/******************************************************************************/
 /**
   Decrypts a block of text with SEED
   @param ct The input ciphertext (16 bytes)
   @param pt The output plaintext (16 bytes)
   @param skey The key as scheduled
 */
-void psSeedDecryptBlock(const unsigned char *ct, unsigned char *pt,
-						psSeedKey_t *skey)
+static void psSeedDecryptBlock(const unsigned char *ct, unsigned char *pt,
+								psSeedKey_t *skey)
 {
 	uint32 P[4];
 	LOAD32H(P[0], ct);
@@ -425,10 +395,9 @@ void psSeedDecryptBlock(const unsigned char *ct, unsigned char *pt,
 	STORE32H(P[3], pt+4);
 	STORE32H(P[0], pt+8);
 	STORE32H(P[1], pt+12);
-	return;
 }
 
+#endif /* USE_MATRIX_SEED */
 
-#endif /* USE_SEED */
 /******************************************************************************/
 
