@@ -1,10 +1,10 @@
 /**
  *	@file    throughputTest.c
- *	@version 33ef80f (HEAD, tag: MATRIXSSL-3-7-2-OPEN, tag: MATRIXSSL-3-7-2-COMM, origin/master, origin/HEAD, master)
+ *	@version 7f5c362 (HEAD -> master)
  *
  */
 /*
- *	Copyright (c) 2013-2015 INSIDE Secure Corporation
+ *	Copyright (c) 2013-2016 INSIDE Secure Corporation
  *	Copyright (c) PeerSec Networks, 2002-2011
  *	All Rights Reserved
  *
@@ -18,7 +18,7 @@
  *	This General Public License does NOT permit incorporating this software
  *	into proprietary programs.  If you are unable to comply with the GPL, a
  *	commercial license for this software may be purchased from INSIDE at
- *	http://www.insidesecure.com/eng/Company/Locations
+ *	http://www.insidesecure.com/
  *
  *	This program is distributed in WITHOUT ANY WARRANTY; without even the
  *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
@@ -49,26 +49,33 @@ static unsigned char key[32] = {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
 			0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
 			0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f };
 
-#define AES_ALG		1
-#define ARC4_ALG	2
-#define DES3_ALG	3
-#define SEED_ALG	4
-#define AES_GCM_ALG	5
-#define IDEA_ALG	6
+enum {
+	AES_ENC_ALG = 1,
+	AES_DEC_ALG,
+	AES_GCM_ALG,
+	ARC4_ALG,
+	DES3_ALG,
+	SEED_ALG,
+	IDEA_ALG,
 
+	AES_HMAC_ALG,
+	AES_HMAC256_ALG,
 
-#define SHA1_ALG	1
-#define SHA224_ALG	2
-#define SHA256_ALG	3
-#define SHA384_ALG	4
-#define SHA512_ALG	5
-#define MD5_ALG		6
+	SHA1_ALG,
+	SHA256_ALG,
+	SHA384_ALG,
+	SHA512_ALG,
+	MD5_ALG
+};
 
-static void runTime(psCipherContext_t *ctx, int32 chunk, int32 alg)
+#if defined(USE_HMAC_SHA1) || defined(USE_HMAC_SHA256)
+static void runWithHmac(psCipherContext_t *ctx, psHmac_t *hmac,
+				int32 hashSize, int32 chunk, int32 alg)
 {
 	psTime_t			start, end;
 	unsigned char		*dataChunk;
 	int32				bytesSent, bytesToSend, round;
+	unsigned char		mac[MAX_HASH_SIZE];
 #ifdef USE_HIGHRES_TIME
 	int32				mod;
 	int64				diffu;
@@ -81,80 +88,161 @@ static void runTime(psCipherContext_t *ctx, int32 chunk, int32 alg)
 	bytesToSend = (DATABYTES_AMOUNT / chunk) * chunk;
 	bytesSent = 0;
 
-	if (alg == AES_ALG) {
+	switch (alg) {
 #ifdef USE_AES
+#ifdef USE_HMAC_SHA1
+	case AES_HMAC_ALG:
 		psGetTime(&start, NULL);
 		while (bytesSent < bytesToSend) {
-			if (psAesEncrypt(ctx, dataChunk, dataChunk, chunk) < 0)  {
-				printf("AES-CBC encrypt error\n");
-				return;
-			}
-			bytesSent += chunk;
-		}
-		psGetTime(&end, NULL);
+#ifdef USE_HMAC_TLS
+			static unsigned char hmacKey[64] = { 0, };
+			unsigned char mac[20];
+			psHmacSha1Tls(hmacKey, 20, dataChunk, chunk,
+						  NULL, 0, NULL, 0, 0, mac);
+#else
+			psHmacSha1Update(&hmac->u.sha1, dataChunk, chunk);
 #endif
-#ifdef USE_ARC4
-	} else if (alg == ARC4_ALG) {
-		psGetTime(&start, NULL);
-		while (bytesSent < bytesToSend) {
-			if (psArc4(ctx, dataChunk, dataChunk, chunk) < 0)  {
-				printf("ARC4 encrypt error\n");
-				return;
-			}
+			psAesEncryptCBC(&ctx->aes, dataChunk, dataChunk, chunk);
 			bytesSent += chunk;
 		}
+		psHmacSha1Final(&hmac->u.sha1, mac);
 		psGetTime(&end, NULL);
+		break;
 #endif
-#ifdef USE_3DES
-	} else if (alg == DES3_ALG) {
+#ifdef USE_HMAC_SHA256
+	case AES_HMAC256_ALG:
 		psGetTime(&start, NULL);
 		while (bytesSent < bytesToSend) {
-			if (psDes3Encrypt(ctx, dataChunk, dataChunk, chunk) < 0)  {
-				printf("3DES encrypt error\n");
-				return;
-			}
+#ifdef USE_HMAC_TLS
+			static unsigned char hmacKey[64] = { 0, };
+			unsigned char mac[32];
+			psHmacSha2Tls(hmacKey, 32, dataChunk, chunk,
+						  NULL, 0, NULL, 0, 0, mac, 32);
+#else
+			psHmacSha256Update(&hmac->u.sha256, dataChunk, chunk);
+#endif
+			psAesEncryptCBC(&ctx->aes, dataChunk, dataChunk, chunk);
+			bytesSent += chunk;
+		}
+		psHmacSha256Final(&hmac->u.sha256, mac);
+		psGetTime(&end, NULL);
+		break;
+#endif
+#endif
+	default:
+		printf("Skipping HMAC Test\n");
+		psFree(dataChunk, NULL);
+		return;
+	}
+
+#ifdef USE_HIGHRES_TIME
+	diffu = psDiffUsecs(start, end);
+	round = (bytesToSend / diffu);
+	mod = (bytesToSend % diffu);
+	printf("%d byte chunks in %lld usecs total for rate of %d.%d MB/sec\n",
+		chunk, (unsigned long long)diffu, round, mod);
+#else
+	diffm = psDiffMsecs(start, end, NULL);
+	round = (bytesToSend / diffm) / 1000;
+	printf("%d byte chunks in %d msecs total for rate of %d MB/sec\n",
+		chunk, diffm, round);
+#endif
+	psFree(dataChunk, NULL);
+}
+#endif /* USE_HMAC */
+
+static void runTime(psCipherContext_t *ctx, psCipherGivContext_t *ctx_giv,
+					int32 chunk, int32 alg)
+{
+	psTime_t			start, end;
+	unsigned char		*dataChunk;
+	int32				bytesSent, bytesToSend, round;
+#ifdef USE_HIGHRES_TIME
+	int32				mod;
+	int64				diffu;
+#else
+	int32				diffm;
+#endif
+
+	dataChunk = psMalloc(NULL, chunk + 16);
+	memset(dataChunk, 0x0, chunk);
+	bytesToSend = (DATABYTES_AMOUNT / chunk) * chunk;
+	bytesSent = 0;
+
+	switch(alg) {
+#ifdef USE_AES_CBC
+	case AES_ENC_ALG:
+		printf("Encrypt ");
+		psGetTime(&start, NULL);
+		while (bytesSent < bytesToSend) {
+			psAesEncryptCBC(&ctx->aes, dataChunk, dataChunk, chunk);
 			bytesSent += chunk;
 		}
 		psGetTime(&end, NULL);
+		break;
+	case AES_DEC_ALG:
+		printf("Decrypt ");
+		psGetTime(&start, NULL);
+		while (bytesSent < bytesToSend) {
+			psAesDecryptCBC(&ctx->aes, dataChunk, dataChunk, chunk);
+			bytesSent += chunk;
+		}
+		psGetTime(&end, NULL);
+		break;
 #endif
 #ifdef USE_AES_GCM
-	} else if (alg == AES_GCM_ALG) {
+	case AES_GCM_ALG:
 		psGetTime(&start, NULL);
 		while (bytesSent < bytesToSend) {
-			if (psAesEncryptGCM(ctx, dataChunk, dataChunk, chunk) < 0)  {
-				printf("SEED encrypt error\n");
-				return;
-			}
+			psAesEncryptGCM(&ctx->aesgcm, dataChunk, dataChunk, chunk);
 			bytesSent += chunk;
 		}
-		psAesGetGCMTag(ctx, 16, dataChunk);
+		psAesGetGCMTag(&ctx->aesgcm, 16, dataChunk);
 		psGetTime(&end, NULL);
+		break;
+#endif
+#ifdef USE_ARC4
+	case ARC4_ALG:
+		psGetTime(&start, NULL);
+		while (bytesSent < bytesToSend) {
+			psArc4(&ctx->arc4, dataChunk, dataChunk, chunk);
+			bytesSent += chunk;
+		}
+		psGetTime(&end, NULL);
+		break;
+#endif
+#ifdef USE_3DES
+	case DES3_ALG:
+		psGetTime(&start, NULL);
+		while (bytesSent < bytesToSend) {
+			psDes3Encrypt(&ctx->des3, dataChunk, dataChunk, chunk);
+			bytesSent += chunk;
+		}
+		psGetTime(&end, NULL);
+		break;
 #endif
 #ifdef USE_SEED
-	} else if (alg == SEED_ALG) {
+	case SEED_ALG:
 		psGetTime(&start, NULL);
 		while (bytesSent < bytesToSend) {
-			if (psSeedEncrypt(ctx, dataChunk, dataChunk, chunk) < 0)  {
-				printf("SEED encrypt error\n");
-				return;
-			}
+			psSeedEncrypt(&ctx->seed, dataChunk, dataChunk, chunk);
 			bytesSent += chunk;
 		}
 		psGetTime(&end, NULL);
+		break;
 #endif
 #ifdef USE_IDEA
-	} else if (alg == IDEA_ALG) {
+	case IDEA_ALG:
 		psGetTime(&start, NULL);
 		while (bytesSent < bytesToSend) {
-			if (psIdeaEncrypt(ctx, dataChunk, dataChunk, chunk) < 0)  {
-				printf("IDEA encrypt error\n");
-				return;
-			}
+			psIdeaEncrypt(&ctx->idea, dataChunk, dataChunk, chunk);
 			bytesSent += chunk;
 		}
 		psGetTime(&end, NULL);
+		break;
 #endif
-	} else {
+	default:
+		psFree(dataChunk, NULL);
 		return;
 	}
 
@@ -165,7 +253,7 @@ static void runTime(psCipherContext_t *ctx, int32 chunk, int32 alg)
 	round = (bytesToSend / diffu);
 	mod = (bytesToSend % diffu);
 	printf("%d byte chunks in %lld usecs total for rate of %d.%d MB/sec\n",
-		chunk, diffu, round, mod);
+		chunk, (unsigned long long)diffu, round, mod);
 #else
 	diffm = psDiffMsecs(start, end, NULL);
 	round = (bytesToSend / diffm) / 1000;
@@ -176,84 +264,195 @@ static void runTime(psCipherContext_t *ctx, int32 chunk, int32 alg)
 }
 
 /******************************************************************************/
-#ifdef USE_AES
+#ifdef USE_AES_CBC
 static int32 psAesTestCBC(void)
 {
 	int32				err;
 	psCipherContext_t	eCtx;
 
-#if !defined PS_AES_IMPROVE_PERF_INCREASE_CODESIZE && !defined USE_AESNI_CRYPTO
+#if defined(USE_MATRIX_AES_CBC) && !defined(PS_AES_IMPROVE_PERF_INCREASE_CODESIZE)
 	_psTrace("##########\n#\n# ");
 	_psTrace("AES speeds can be improved by enabling\n# ");
 	_psTrace("PS_AES_IMPROVE_PERF_INCREASE_CODESIZE in cryptoConfig.h\n");
 	_psTrace("#\n#\n#########\n");
 #endif
+
 	_psTrace("***** AES-128 CBC *****\n");
-	if ((err = psAesInit(&eCtx, iv, key, 16)) != PS_SUCCESS) {
-		_psTraceInt("FAILED:  psAesInit returned %d\n", err);
+	if ((err = psAesInitCBC(&eCtx.aes, iv, key, 16, PS_AES_ENCRYPT)) != PS_SUCCESS) {
+		_psTraceInt("FAILED:  returned %d\n", err);
  		return err;
 	}
-	runTime(&eCtx, TINY_CHUNKS, AES_ALG);
-	runTime(&eCtx, SMALL_CHUNKS, AES_ALG);
-	runTime(&eCtx, MEDIUM_CHUNKS, AES_ALG);
-	runTime(&eCtx, LARGE_CHUNKS, AES_ALG);
-	runTime(&eCtx, HUGE_CHUNKS, AES_ALG);
-#ifndef USE_AESNI_CRYPTO
+	runTime(&eCtx, NULL, TINY_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, SMALL_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, MEDIUM_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, LARGE_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, LARGE_CHUNKS, AES_DEC_ALG);
+	runTime(&eCtx, NULL, HUGE_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, HUGE_CHUNKS, AES_DEC_ALG);
+	psAesClearCBC(&eCtx.aes);
+
 	_psTrace("***** AES-192 CBC *****\n");
-	if ((err = psAesInit(&eCtx, iv, key, 24)) != PS_SUCCESS) {
-		_psTraceInt("FAILED:  psAesInit returned %d\n", err);
+	if ((err = psAesInitCBC(&eCtx.aes, iv, key, 24, PS_AES_ENCRYPT)) != PS_SUCCESS) {
+		_psTraceInt("FAILED:  returned %d\n", err);
  		return err;
 	}
-	runTime(&eCtx, TINY_CHUNKS, AES_ALG);
-	runTime(&eCtx, SMALL_CHUNKS, AES_ALG);
-	runTime(&eCtx, MEDIUM_CHUNKS, AES_ALG);
-	runTime(&eCtx, LARGE_CHUNKS, AES_ALG);
-	runTime(&eCtx, HUGE_CHUNKS, AES_ALG);
-#endif
+	runTime(&eCtx, NULL, TINY_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, SMALL_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, MEDIUM_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, LARGE_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, LARGE_CHUNKS, AES_DEC_ALG);
+	runTime(&eCtx, NULL, HUGE_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, HUGE_CHUNKS, AES_DEC_ALG);
+	psAesClearCBC(&eCtx.aes);
+
 	_psTrace("***** AES-256 CBC *****\n");
-	if ((err = psAesInit(&eCtx, iv, key, 32)) != PS_SUCCESS) {
-		_psTraceInt("FAILED:  psAesInit returned %d\n", err);
+	if ((err = psAesInitCBC(&eCtx.aes, iv, key, 32, PS_AES_ENCRYPT)) != PS_SUCCESS) {
+		_psTraceInt("FAILED:  returned %d\n", err);
  		return err;
 	}
-	runTime(&eCtx, TINY_CHUNKS, AES_ALG);
-	runTime(&eCtx, SMALL_CHUNKS, AES_ALG);
-	runTime(&eCtx, MEDIUM_CHUNKS, AES_ALG);
-	runTime(&eCtx, LARGE_CHUNKS, AES_ALG);
-	runTime(&eCtx, HUGE_CHUNKS, AES_ALG);
+	runTime(&eCtx, NULL, TINY_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, SMALL_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, MEDIUM_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, LARGE_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, LARGE_CHUNKS, AES_DEC_ALG);
+	runTime(&eCtx, NULL, HUGE_CHUNKS, AES_ENC_ALG);
+	runTime(&eCtx, NULL, HUGE_CHUNKS, AES_DEC_ALG);
+	psAesClearCBC(&eCtx.aes);
 
 	return 0;
 }
 
+#if defined(USE_HMAC_SHA1) || defined(USE_HMAC_SHA256)
+static int32 psAesTestCBCHmac(void)
+{
+	int32				err;
+	psCipherContext_t	eCtx;
+	psHmac_t			hCtx;
+
+#if defined(USE_MATRIX_AES_CBC) && !defined(PS_AES_IMPROVE_PERF_INCREASE_CODESIZE)
+	_psTrace("##########\n#\n# ");
+	_psTrace("AES speeds can be improved by enabling\n# ");
+	_psTrace("PS_AES_IMPROVE_PERF_INCREASE_CODESIZE in cryptoConfig.h\n");
+	_psTrace("#\n#\n#########\n");
+#endif
+
+#ifdef USE_HMAC_SHA1
+	_psTrace("***** AES-128 CBC + SHA1-HMAC *****\n");
+	if ((err = psAesInitCBC(&eCtx.aes, iv, key, 16, PS_AES_ENCRYPT)) != PS_SUCCESS) {
+		_psTraceInt("FAILED:  returned %d\n", err);
+ 		return err;
+	}
+	psHmacSha1Init(&hCtx.u.sha1, key, SHA1_HASH_SIZE);
+	runWithHmac(&eCtx, &hCtx, 0, TINY_CHUNKS, AES_HMAC_ALG);
+	psHmacSha1Init(&hCtx.u.sha1, key, SHA1_HASH_SIZE);
+	runWithHmac(&eCtx, &hCtx, 0, SMALL_CHUNKS, AES_HMAC_ALG);
+	psHmacSha1Init(&hCtx.u.sha1, key, SHA1_HASH_SIZE);
+	runWithHmac(&eCtx, &hCtx, 0, MEDIUM_CHUNKS, AES_HMAC_ALG);
+	psHmacSha1Init(&hCtx.u.sha1, key, SHA1_HASH_SIZE);
+	runWithHmac(&eCtx, &hCtx, 0, LARGE_CHUNKS, AES_HMAC_ALG);
+	psHmacSha1Init(&hCtx.u.sha1, key, SHA1_HASH_SIZE);
+	runWithHmac(&eCtx, &hCtx, 0, HUGE_CHUNKS, AES_HMAC_ALG);
+	psAesClearCBC(&eCtx.aes);
+
+	_psTrace("***** AES-256 CBC + SHA1-HMAC *****\n");
+	if ((err = psAesInitCBC(&eCtx.aes, iv, key, 32, PS_AES_ENCRYPT)) != PS_SUCCESS) {
+		_psTraceInt("FAILED:  returned %d\n", err);
+ 		return err;
+	}
+	psHmacSha1Init(&hCtx.u.sha1, key, SHA1_HASH_SIZE);
+	runWithHmac(&eCtx, &hCtx, 0, TINY_CHUNKS, AES_HMAC_ALG);
+	psHmacSha1Init(&hCtx.u.sha1, key, SHA1_HASH_SIZE);
+	runWithHmac(&eCtx, &hCtx, 0, SMALL_CHUNKS, AES_HMAC_ALG);
+	psHmacSha1Init(&hCtx.u.sha1, key, SHA1_HASH_SIZE);
+	runWithHmac(&eCtx, &hCtx, 0, MEDIUM_CHUNKS, AES_HMAC_ALG);
+	psHmacSha1Init(&hCtx.u.sha1, key, SHA1_HASH_SIZE);
+	runWithHmac(&eCtx, &hCtx, 0, LARGE_CHUNKS, AES_HMAC_ALG);
+	psHmacSha1Init(&hCtx.u.sha1, key, SHA1_HASH_SIZE);
+	runWithHmac(&eCtx, &hCtx, 0, HUGE_CHUNKS, AES_HMAC_ALG);
+	psAesClearCBC(&eCtx.aes);
+#endif
+
+#ifdef USE_HMAC_SHA256
+	_psTrace("***** AES-128 CBC + SHA256-HMAC *****\n");
+	if ((err = psAesInitCBC(&eCtx.aes, iv, key, 16, PS_AES_ENCRYPT)) != PS_SUCCESS) {
+		_psTraceInt("FAILED:  returned %d\n", err);
+ 		return err;
+	}
+	psHmacSha256Init(&hCtx.u.sha256, key, 32);
+	runWithHmac(&eCtx, &hCtx, SHA256_HASH_SIZE, TINY_CHUNKS, AES_HMAC256_ALG);
+	psHmacSha256Init(&hCtx.u.sha256, key, 32);
+	runWithHmac(&eCtx, &hCtx, SHA256_HASH_SIZE, SMALL_CHUNKS, AES_HMAC256_ALG);
+	psHmacSha256Init(&hCtx.u.sha256, key, 32);
+	runWithHmac(&eCtx, &hCtx, SHA256_HASH_SIZE, MEDIUM_CHUNKS, AES_HMAC256_ALG);
+	psHmacSha256Init(&hCtx.u.sha256, key, 32);
+	runWithHmac(&eCtx, &hCtx, SHA256_HASH_SIZE, LARGE_CHUNKS, AES_HMAC256_ALG);
+	psHmacSha256Init(&hCtx.u.sha256, key, 32);
+	runWithHmac(&eCtx, &hCtx, SHA256_HASH_SIZE, HUGE_CHUNKS, AES_HMAC256_ALG);
+	psAesClearCBC(&eCtx.aes);
+
+	_psTrace("***** AES-256 CBC + SHA256-HMAC *****\n");
+	if ((err = psAesInitCBC(&eCtx.aes, iv, key, 32, PS_AES_ENCRYPT)) != PS_SUCCESS) {
+		_psTraceInt("FAILED:  returned %d\n", err);
+ 		return err;
+	}
+	psHmacSha256Init(&hCtx.u.sha256, key, 32);
+	runWithHmac(&eCtx, &hCtx, SHA256_HASH_SIZE, TINY_CHUNKS, AES_HMAC256_ALG);
+	psHmacSha256Init(&hCtx.u.sha256, key, 32);
+	runWithHmac(&eCtx, &hCtx, SHA256_HASH_SIZE, SMALL_CHUNKS, AES_HMAC256_ALG);
+	psHmacSha256Init(&hCtx.u.sha256, key, 32);
+	runWithHmac(&eCtx, &hCtx, SHA256_HASH_SIZE, MEDIUM_CHUNKS, AES_HMAC256_ALG);
+	psHmacSha256Init(&hCtx.u.sha256, key, 32);
+	runWithHmac(&eCtx, &hCtx, SHA256_HASH_SIZE, LARGE_CHUNKS, AES_HMAC256_ALG);
+	psHmacSha256Init(&hCtx.u.sha256, key, 32);
+	runWithHmac(&eCtx, &hCtx, SHA256_HASH_SIZE, HUGE_CHUNKS, AES_HMAC256_ALG);
+	psHmacSha256Init(&hCtx.u.sha256, key, 32);
+	psAesClearCBC(&eCtx.aes);
+#endif
+
+	return 0;
+}
+#endif /* USE_HMAC */
+
+/******************************************************************************/
+
 #ifdef USE_AES_GCM
 int32 psAesTestGCM(void)
 {
 	int32				err;
-	psCipherContext_t	eCtx;
+    psCipherContext_t	eCtx;
+	psCipherGivContext_t eCtxGiv;
 
+	memset(&eCtxGiv, 0, sizeof(eCtxGiv));
 
+#ifndef USE_LIBSODIUM_AES_GCM
 	_psTrace("***** AES-GCM-128 *****\n");
-	if ((err = psAesInitGCM(&eCtx, key, 16)) != PS_SUCCESS) {
+	if ((err = psAesInitGCM(&eCtx.aesgcm, key, 16)) != PS_SUCCESS) {
 		_psTraceInt("FAILED:  psAesInitGCM returned %d\n", err);
  		return err;
 	}
-	psAesReadyGCM(&eCtx, iv, iv, 16);
-	runTime(&eCtx, TINY_CHUNKS, AES_GCM_ALG);
-	runTime(&eCtx, SMALL_CHUNKS, AES_GCM_ALG);
-	runTime(&eCtx, MEDIUM_CHUNKS, AES_GCM_ALG);
-	runTime(&eCtx, LARGE_CHUNKS, AES_GCM_ALG);
-	runTime(&eCtx, HUGE_CHUNKS, AES_GCM_ALG);
+	psAesReadyGCM(&eCtx.aesgcm, iv, iv, 16);
+	runTime(&eCtx, &eCtxGiv, TINY_CHUNKS, AES_GCM_ALG);
+	runTime(&eCtx, &eCtxGiv, SMALL_CHUNKS, AES_GCM_ALG);
+	runTime(&eCtx, &eCtxGiv, MEDIUM_CHUNKS, AES_GCM_ALG);
+	runTime(&eCtx, &eCtxGiv, LARGE_CHUNKS, AES_GCM_ALG);
+	runTime(&eCtx, &eCtxGiv, HUGE_CHUNKS, AES_GCM_ALG);
+#else
+	_psTrace("***** Skipping AES-GCM-128 *****\n");
+#endif /* !USE_LIBSODIUM */
 
 	_psTrace("***** AES-GCM-256 *****\n");
-	if ((err = psAesInitGCM(&eCtx, key, 32)) != PS_SUCCESS) {
+	if ((err = psAesInitGCM(&eCtx.aesgcm, key, 32)) != PS_SUCCESS) {
 		_psTraceInt("FAILED:  psAesInitGCM returned %d\n", err);
  		return err;
 	}
-	psAesReadyGCM(&eCtx, iv, iv, 16);
-	runTime(&eCtx, TINY_CHUNKS, AES_GCM_ALG);
-	runTime(&eCtx, SMALL_CHUNKS, AES_GCM_ALG);
-	runTime(&eCtx, MEDIUM_CHUNKS, AES_GCM_ALG);
-	runTime(&eCtx, LARGE_CHUNKS, AES_GCM_ALG);
-	runTime(&eCtx, HUGE_CHUNKS, AES_GCM_ALG);
+	psAesReadyGCM(&eCtx.aesgcm, iv, iv, 16);
+	runTime(&eCtx, &eCtxGiv, TINY_CHUNKS, AES_GCM_ALG);
+	runTime(&eCtx, &eCtxGiv, SMALL_CHUNKS, AES_GCM_ALG);
+	runTime(&eCtx, &eCtxGiv, MEDIUM_CHUNKS, AES_GCM_ALG);
+	runTime(&eCtx, &eCtxGiv, LARGE_CHUNKS, AES_GCM_ALG);
+	runTime(&eCtx, &eCtxGiv, HUGE_CHUNKS, AES_GCM_ALG);
+
+	psAesClearGCM(&eCtx.aesgcm);
 
 	return PS_SUCCESS;
 }
@@ -267,20 +466,23 @@ int32 psDes3Test(void)
 {
 	psCipherContext_t	eCtx;
 
-#ifndef PS_3DES_IMPROVE_PERF_INCREASE_CODESIZE
+#if defined(USE_MATRIX_3DES) && !defined(PS_3DES_IMPROVE_PERF_INCREASE_CODESIZE)
 	_psTrace("##########\n#\n# ");
 	_psTrace("3DES speeds can be improved by enabling\n# ");
 	_psTrace("PS_3DES_IMPROVE_PERF_INCREASE_CODESIZE in cryptoConfig.h\n");
 	_psTrace("#\n#\n#########\n");
 #endif
 
-	psDes3Init(&eCtx, iv, key, 24);
+	psDes3Init(&eCtx.des3, iv, key);
+
+	runTime(&eCtx, NULL, TINY_CHUNKS, DES3_ALG);
+	runTime(&eCtx, NULL, SMALL_CHUNKS, DES3_ALG);
+	runTime(&eCtx, NULL, MEDIUM_CHUNKS, DES3_ALG);
+	runTime(&eCtx, NULL, LARGE_CHUNKS, DES3_ALG);
+	runTime(&eCtx, NULL, HUGE_CHUNKS, DES3_ALG);
+
+	psDes3Clear(&eCtx.des3);
 
-	runTime(&eCtx, TINY_CHUNKS, DES3_ALG);
-	runTime(&eCtx, SMALL_CHUNKS, DES3_ALG);
-	runTime(&eCtx, MEDIUM_CHUNKS, DES3_ALG);
-	runTime(&eCtx, LARGE_CHUNKS, DES3_ALG);
-	runTime(&eCtx, HUGE_CHUNKS, DES3_ALG);
 	return 0;
 }
 #endif /* USE_3DES */
@@ -291,13 +493,15 @@ int32 psArc4Test(void)
 {
 	psCipherContext_t	eCtx;
 
-	psArc4Init(&eCtx, key, 16);
+	psArc4Init(&eCtx.arc4, key, 16);
 
-	runTime(&eCtx, TINY_CHUNKS, ARC4_ALG);
-	runTime(&eCtx, SMALL_CHUNKS, ARC4_ALG);
-	runTime(&eCtx, MEDIUM_CHUNKS, ARC4_ALG);
-	runTime(&eCtx, LARGE_CHUNKS, ARC4_ALG);
-	runTime(&eCtx, HUGE_CHUNKS, ARC4_ALG);
+	runTime(&eCtx, NULL, TINY_CHUNKS, ARC4_ALG);
+	runTime(&eCtx, NULL, SMALL_CHUNKS, ARC4_ALG);
+	runTime(&eCtx, NULL, MEDIUM_CHUNKS, ARC4_ALG);
+	runTime(&eCtx, NULL, LARGE_CHUNKS, ARC4_ALG);
+	runTime(&eCtx, NULL, HUGE_CHUNKS, ARC4_ALG);
+
+	psArc4Clear(&eCtx.arc4);
 
 	return 0;
 }
@@ -310,13 +514,15 @@ int32 psSeedTest(void)
 {
 	psCipherContext_t	eCtx;
 
-	psSeedInit(&eCtx, iv, key, 16);
+	psSeedInit(&eCtx.seed, iv, key);
 
-	runTime(&eCtx, TINY_CHUNKS, SEED_ALG);
-	runTime(&eCtx, SMALL_CHUNKS, SEED_ALG);
-	runTime(&eCtx, MEDIUM_CHUNKS, SEED_ALG);
-	runTime(&eCtx, LARGE_CHUNKS, SEED_ALG);
-	runTime(&eCtx, HUGE_CHUNKS, SEED_ALG);
+	runTime(&eCtx, NULL, TINY_CHUNKS, SEED_ALG);
+	runTime(&eCtx, NULL, SMALL_CHUNKS, SEED_ALG);
+	runTime(&eCtx, NULL, MEDIUM_CHUNKS, SEED_ALG);
+	runTime(&eCtx, NULL, LARGE_CHUNKS, SEED_ALG);
+	runTime(&eCtx, NULL, HUGE_CHUNKS, SEED_ALG);
+
+	psSeedClear(&eCtx.seed);
 
 	return PS_SUCCESS;
 }
@@ -327,17 +533,19 @@ int32 psIdeaTest(void)
 {
 	psCipherContext_t	eCtx;
 
-	psIdeaInit(&eCtx, iv, key, 16);
+	psIdeaInit(&eCtx.idea, iv, key);
 
-	runTime(&eCtx, TINY_CHUNKS, IDEA_ALG);
-	runTime(&eCtx, SMALL_CHUNKS, IDEA_ALG);
-	runTime(&eCtx, MEDIUM_CHUNKS, IDEA_ALG);
-	runTime(&eCtx, LARGE_CHUNKS, IDEA_ALG);
-	runTime(&eCtx, HUGE_CHUNKS, IDEA_ALG);
+	runTime(&eCtx, NULL, TINY_CHUNKS, IDEA_ALG);
+	runTime(&eCtx, NULL, SMALL_CHUNKS, IDEA_ALG);
+	runTime(&eCtx, NULL, MEDIUM_CHUNKS, IDEA_ALG);
+	runTime(&eCtx, NULL, LARGE_CHUNKS, IDEA_ALG);
+	runTime(&eCtx, NULL, HUGE_CHUNKS, IDEA_ALG);
+
+	psIdeaClear(&eCtx.idea);
 
 	return PS_SUCCESS;
 }
-#endif /* USE_SEED */
+#endif /* USE_IDEA */
 /******************************************************************************/
 
 void runDigestTime(psDigestContext_t *ctx, int32 chunk, int32 alg)
@@ -357,67 +565,65 @@ void runDigestTime(psDigestContext_t *ctx, int32 chunk, int32 alg)
 	bytesToSend = (DATABYTES_AMOUNT / chunk) * chunk;
 	bytesSent = 0;
 
-	if (alg == SHA1_ALG) {
+	switch (alg) {
 #ifdef USE_SHA1
+	case SHA1_ALG:
 		psGetTime(&start, NULL);
 		while (bytesSent < bytesToSend) {
-			psSha1Update(ctx, dataChunk, chunk);
+			psSha1Update(&ctx->sha1, dataChunk, chunk);
 			bytesSent += chunk;
 		}
-		psSha1Final(ctx, hashout);
-		psGetTime(&end, NULL);
-#endif
-#ifdef USE_SHA224
-	} else if (alg == SHA224_ALG) {
-		psGetTime(&start, NULL);
-		while (bytesSent < bytesToSend) {
-			psSha224Update(ctx, dataChunk, chunk);
-			bytesSent += chunk;
-		}
-		psSha224Final(ctx, hashout);
+		psSha1Final(&ctx->sha1, hashout);
 		psGetTime(&end, NULL);
+		break;
 #endif
 #ifdef USE_SHA256
-	} else if (alg == SHA256_ALG) {
+	case SHA256_ALG:
 		psGetTime(&start, NULL);
 		while (bytesSent < bytesToSend) {
-			psSha256Update(ctx, dataChunk, chunk);
+			psSha256Update(&ctx->sha256, dataChunk, chunk);
 			bytesSent += chunk;
 		}
-		psSha256Final(ctx, hashout);
+		psSha256Final(&ctx->sha256, hashout);
 		psGetTime(&end, NULL);
+		break;
 #endif
 #ifdef USE_SHA384
-	} else if (alg == SHA384_ALG) {
+	case SHA384_ALG:
 		psGetTime(&start, NULL);
 		while (bytesSent < bytesToSend) {
-			psSha384Update(ctx, dataChunk, chunk);
+			psSha384Update(&ctx->sha384, dataChunk, chunk);
 			bytesSent += chunk;
 		}
-		psSha384Final(ctx, hashout);
+		psSha384Final(&ctx->sha384, hashout);
 		psGetTime(&end, NULL);
+		break;
 #endif
 #ifdef USE_SHA512
-	} else if (alg == SHA512_ALG) {
+	case SHA512_ALG:
 		psGetTime(&start, NULL);
 		while (bytesSent < bytesToSend) {
-			psSha512Update(ctx, dataChunk, chunk);
+			psSha512Update(&ctx->sha512, dataChunk, chunk);
 			bytesSent += chunk;
 		}
-		psSha512Final(ctx, hashout);
+		psSha512Final(&ctx->sha512, hashout);
 		psGetTime(&end, NULL);
+		break;
 #endif
 #ifdef USE_MD5
-	} else if (alg == MD5_ALG) {
+	case MD5_ALG:
 		psGetTime(&start, NULL);
 		while (bytesSent < bytesToSend) {
-			psMd5Update(ctx, dataChunk, chunk);
+			psMd5Update(&ctx->md5, dataChunk, chunk);
 			bytesSent += chunk;
 		}
-		psMd5Final(ctx, hashout);
+		psMd5Final(&ctx->md5, hashout);
 		psGetTime(&end, NULL);
+		break;
 #endif
-	} else {
+	default:
+		psFree(dataChunk, NULL);
+		printf("Skipping Digest Tests\n");
 		return;
 	}
 
@@ -426,36 +632,30 @@ void runDigestTime(psDigestContext_t *ctx, int32 chunk, int32 alg)
 	round = (bytesToSend / diffu);
 	mod = (bytesToSend % diffu);
 	printf("%d byte chunks in %lld usecs total for rate of %d.%d MB/sec\n",
-		chunk, diffu, round, mod);
+		chunk, (unsigned long long)diffu, round, mod);
 #else
 	diffm = psDiffMsecs(start, end, NULL);
 	round = (bytesToSend / diffm) / 1000;
 	printf("%d byte chunks in %d msecs total for rate of %d MB/sec\n",
 		chunk, diffm, round);
 #endif
-
+	psFree(dataChunk, NULL);
 }
 
-
 /******************************************************************************/
 #ifdef USE_SHA1
-#ifndef PS_SHA1_IMPROVE_PERF_INCREASE_CODESIZE
-static int sha1faster = 1;
-#else
-static int sha1faster = 0;
-#endif
 int32  psSha1Test(void)
 {
 	psDigestContext_t	ctx;
+#if defined(USE_MATRIX_SHA1) && !defined(PS_SHA1_IMPROVE_PERF_INCREASE_CODESIZE)
+	_psTrace("##########\n#\n# ");
+	_psTrace("SHA-1 speeds can be improved by enabling\n# ");
+	_psTrace("PS_SHA1_IMPROVE_PERF_INCREASE_CODESIZE in cryptoConfig.h\n");
+	_psTrace("#\n#\n#########\n");
+#endif
 
-	if (sha1faster) {
-		_psTrace("##########\n#\n# ");
-		_psTrace("SHA-1 speeds can be improved by enabling\n# ");
-		_psTrace("PS_SHA1_IMPROVE_PERF_INCREASE_CODESIZE in cryptoConfig.h\n");
-		_psTrace("#\n#\n#########\n");
-	}
 
-	psSha1Init(&ctx);
+	psSha1Init(&ctx.sha1);
 	runDigestTime(&ctx, TINY_CHUNKS, SHA1_ALG);
 	runDigestTime(&ctx, SMALL_CHUNKS, SHA1_ALG);
 	runDigestTime(&ctx, MEDIUM_CHUNKS, SHA1_ALG);
@@ -470,23 +670,11 @@ int32  psSha1Test(void)
 
 /******************************************************************************/
 #ifdef USE_SHA256
-#ifndef PS_SHA1_IMPROVE_PERF_INCREASE_CODESIZE
-static int sha256faster = 1;
-#else
-static int sha256faster = 0;
-#endif
 int32 psSha256Test(void)
 {
 	psDigestContext_t	ctx;
 
-	if (sha256faster) {
-		_psTrace("##########\n#\n# ");
-		_psTrace("SHA-256 speeds can be improved by enabling\n# ");
-		_psTrace("PS_SHA256_IMPROVE_PERF_INCREASE_CODESIZE in cryptoConfig.h\n");
-		_psTrace("#\n#\n#########\n");
-	}
-
-	psSha256Init(&ctx);
+	psSha256Init(&ctx.sha256);
 	runDigestTime(&ctx, TINY_CHUNKS, SHA256_ALG);
 	runDigestTime(&ctx, SMALL_CHUNKS, SHA256_ALG);
 	runDigestTime(&ctx, MEDIUM_CHUNKS, SHA256_ALG);
@@ -498,28 +686,12 @@ int32 psSha256Test(void)
 #endif /* USE_SHA256 */
 /******************************************************************************/
 
-#ifdef USE_SHA224
-int32 psSha224Test(void)
-{
-	psDigestContext_t	ctx;
-
-	psSha224Init(&ctx);
-	runDigestTime(&ctx, TINY_CHUNKS, SHA224_ALG);
-	runDigestTime(&ctx, SMALL_CHUNKS, SHA224_ALG);
-	runDigestTime(&ctx, MEDIUM_CHUNKS, SHA224_ALG);
-	runDigestTime(&ctx, LARGE_CHUNKS, SHA224_ALG);
-	runDigestTime(&ctx, HUGE_CHUNKS, SHA224_ALG);
-
-	return PS_SUCCESS;
-}
-#endif /* USE_SHA224 */
-
 #ifdef USE_SHA384
 int32 psSha384Test(void)
 {
 	psDigestContext_t	ctx;
 
-	psSha384Init(&ctx);
+	psSha384Init(&ctx.sha384);
 	runDigestTime(&ctx, TINY_CHUNKS, SHA384_ALG);
 	runDigestTime(&ctx, SMALL_CHUNKS, SHA384_ALG);
 	runDigestTime(&ctx, MEDIUM_CHUNKS, SHA384_ALG);
@@ -532,23 +704,11 @@ int32 psSha384Test(void)
 
 
 #ifdef USE_SHA512
-#ifndef PS_SHA512_IMPROVE_PERF_INCREASE_CODESIZE
-static int sha512faster = 1;
-#else
-static int sha512faster = 0;
-#endif
 int32  psSha512Test(void)
 {
 	psDigestContext_t	ctx;
 
-	if (sha512faster) {
-		_psTrace("##########\n#\n# ");
-		_psTrace("SHA-512 speeds MIGHT improve by enabling\n# ");
-		_psTrace("PS_SHA512_IMPROVE_PERF_INCREASE_CODESIZE in cryptoConfig.h\n");
-		_psTrace("#\n#\n#########\n");
-	}
-
-	psSha512Init(&ctx);
+	psSha512Init(&ctx.sha512);
 	runDigestTime(&ctx, TINY_CHUNKS, SHA512_ALG);
 	runDigestTime(&ctx, SMALL_CHUNKS, SHA512_ALG);
 	runDigestTime(&ctx, MEDIUM_CHUNKS, SHA512_ALG);
@@ -562,23 +722,17 @@ int32  psSha512Test(void)
 
 /******************************************************************************/
 #ifdef USE_MD5
-#ifndef PS_MD5_IMPROVE_PERF_INCREASE_CODESIZE
-static int md5faster = 1;
-#else
-static int md5faster = 0;
-#endif
 int32 psMd5Test(void)
 {
 	psDigestContext_t	ctx;
+#if defined(USE_MATRIX_MD5) && !defined(PS_MD5_IMPROVE_PERF_INCREASE_CODESIZE)
+	_psTrace("##########\n#\n# ");
+	_psTrace("MD5 speeds can be improved by enabling\n# ");
+	_psTrace("PS_MD5_IMPROVE_PERF_INCREASE_CODESIZE in cryptoConfig.h\n");
+	_psTrace("#\n#\n#########\n");
+#endif
 
-	if (md5faster) {
-		_psTrace("##########\n#\n# ");
-		_psTrace("MD5 speeds can be improved by enabling\n# ");
-		_psTrace("PS_MD5_IMPROVE_PERF_INCREASE_CODESIZE in cryptoConfig.h\n");
-		_psTrace("#\n#\n#########\n");
-	}
-
-	psMd5Init(&ctx);
+	psMd5Init(&ctx.md5);
 	runDigestTime(&ctx, TINY_CHUNKS, MD5_ALG);
 	runDigestTime(&ctx, SMALL_CHUNKS, MD5_ALG);
 	runDigestTime(&ctx, MEDIUM_CHUNKS, MD5_ALG);
@@ -618,6 +772,9 @@ typedef struct {
 static test_t tests[] = {
 #ifdef USE_AES
 {psAesTestCBC, "***** AES-CBC TESTS *****"},
+#if defined(USE_HMAC_SHA1) || defined(USE_HMAC_SHA256)
+{psAesTestCBCHmac, "***** AES-CBC + HMAC TESTS *****"},
+#endif
 #ifdef USE_AES_GCM
 {psAesTestGCM, "***** AES-GCM TESTS *****"},
 #endif
@@ -668,13 +825,6 @@ static test_t tests[] = {
 #endif
 , "***** SHA256 TESTS *****"},
 
-#ifdef USE_SHA224
-{psSha224Test
-#else
-{NULL
-#endif
-, "***** SHA224 TESTS *****"},
-
 #ifdef USE_SHA384
 {psSha384Test
 #else
