@@ -1,11 +1,11 @@
 /**
  *	@file    matrixsslApi.c
- *	@version 33ef80f (HEAD, tag: MATRIXSSL-3-7-2-OPEN, tag: MATRIXSSL-3-7-2-COMM, origin/master, origin/HEAD, master)
+ *	@version 7f5c362 (HEAD -> master)
  *
  *	MatrixSSL Public API Layer.
  */
 /*
- *	Copyright (c) 2013-2015 INSIDE Secure Corporation
+ *	Copyright (c) 2013-2016 INSIDE Secure Corporation
  *	Copyright (c) PeerSec Networks, 2002-2011
  *	All Rights Reserved
  *
@@ -19,7 +19,7 @@
  *	This General Public License does NOT permit incorporating this software
  *	into proprietary programs.  If you are unable to comply with the GPL, a
  *	commercial license for this software may be purchased from INSIDE at
- *	http://www.insidesecure.com/eng/Company/Locations
+ *	http://www.insidesecure.com/
  *
  *	This program is distributed in WITHOUT ANY WARRANTY; without even the
  *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
@@ -52,7 +52,7 @@
 			and for subsequent connections should be used without modification
 			of the cipherId.
 
-	cipherSpec	Requested cipher to negotiate to (0 for server's choice)
+	cipherSpec	Array of requested ciphers to negotiate to (0 for server's choice)
 				If non-zero, and server doesn't have it, conn will fail
 
 	certCb	Optional callback to call when validating cert
@@ -68,19 +68,17 @@
 	flags	TODO out of date
 
 	Return	MATRIXSSL_REQUEST_SEND on success
-			< 0 on error
-
-	Do not need to call DeleteSession on failure
+			< 0 on error. Do not need to call DeleteSession on failure
  */
 
 #ifdef USE_CLIENT_SIDE_SSL
-int32 matrixSslNewClientSession(ssl_t **ssl, const sslKeys_t *keys,
-		sslSessionId_t *sid, uint32 cipherSpec[], uint16 cipherSpecLen,
-		int32 (*certCb)(ssl_t *ssl, psX509Cert_t *cert, int32 alert),
-		const char *expectedName,
-		tlsExtension_t *extensions, int32 (*extCb)(ssl_t *ssl,
-		unsigned short extType, unsigned short extLen, void *e),
-		sslSessOpts_t *options)
+int32_t matrixSslNewClientSession(ssl_t **ssl, const sslKeys_t *keys,
+				sslSessionId_t *sid,
+				const uint16_t cipherSpec[], uint8_t cipherSpecLen,
+				sslCertCb_t certCb,
+				const char *expectedName, tlsExtension_t *extensions,
+				sslExtCb_t extCb,
+				sslSessOpts_t *options)
 {
 	ssl_t		*lssl;
 	psBuf_t		tmp;
@@ -100,9 +98,7 @@ int32 matrixSslNewClientSession(ssl_t **ssl, const sslKeys_t *keys,
 	*ssl = NULL;
 	lssl = NULL;
 
-/*
-	Give priority to cipher suite if session id is provided and doesn't match
-*/
+	/* Give priority to cipher suite if session id is provided and doesn't match */
 	if (cipherSpec != NULL && cipherSpec[0] != 0 && sid != NULL &&
 			sid->cipherId != 0) {
 		rc = 1;
@@ -135,12 +131,11 @@ int32 matrixSslNewClientSession(ssl_t **ssl, const sslKeys_t *keys,
 		strcpy(lssl->expectedName, expectedName);
 	}
 	if (certCb) {
-		matrixSslSetCertValidator(lssl, (sslCertCb_t)certCb);
+		matrixSslSetCertValidator(lssl, certCb);
 	}
 #endif
-
 	if (extCb) {
-		lssl->extCb = (sslExtCb_t)extCb;
+		lssl->extCb = extCb;
 	}
 RETRY_HELLO:
 	tmp.size = lssl->outsize;
@@ -196,7 +191,6 @@ void matrixSslClearSessionId(sslSessionId_t *sess)
 
 void matrixSslDeleteSessionId(sslSessionId_t *sess)
 {
-	psPool_t	*pool;
 
 	if (sess == NULL) {
 		return;
@@ -206,10 +200,9 @@ void matrixSslDeleteSessionId(sslSessionId_t *sess)
 		psFree(sess->sessionTicket, sess->pool);
 	}
 #endif
-	pool = sess->pool;
-	memset(sess, 0x0, sizeof(sslSessionId_t));
-	psFree(sess, pool);
 
+	memset(sess, 0x0, sizeof(sslSessionId_t));
+	psFree(sess, NULL);
 }
 
 #endif /* USE_CLIENT_SIDE_SSL */
@@ -224,9 +217,27 @@ void matrixSslDeleteSessionId(sslSessionId_t *sess)
 	Return	MATRIXSSL_SUCCESS on success
 			< 0 on error
  */
+
+int32 matrixSslNewServer(ssl_t **ssl,
+				pubkeyCb_t pubkeyCb, pskCb_t pskCb, sslCertCb_t certCb,
+				sslSessOpts_t *options)
+{
+	int32	rc;
+	
+	if ((rc = matrixSslNewServerSession(ssl, NULL, certCb, options)) < 0) {
+		return rc;
+	}
+	
+	(*ssl)->sec.pskCb = (pskCb_t)pskCb;
+#ifndef USE_ONLY_PSK_CIPHER_SUITE
+	(*ssl)->sec.pubkeyCb = (pubkeyCb_t)pubkeyCb;
+#endif
+	return MATRIXSSL_SUCCESS;
+}
+
 int32 matrixSslNewServerSession(ssl_t **ssl, const sslKeys_t *keys,
-			int32 (*certCb)(ssl_t *ssl, psX509Cert_t *cert, int32 alert),
-			sslSessOpts_t *options)
+				sslCertCb_t certCb,
+				sslSessOpts_t *options)
 {
 	ssl_t		*lssl;
 
@@ -259,15 +270,6 @@ int32 matrixSslNewServerSession(ssl_t **ssl, const sslKeys_t *keys,
 	}
 #endif /* USE_CLIENT_AUTH */
 
-#ifdef USE_ECC
-	/* Must persist.  Set to compiled-in set if not provided by user */
-	if (options->ecFlags <= 0) {
-		lssl->ecFlags = compiledInEcFlags();
-	} else {
-		lssl->ecFlags = options->ecFlags;
-	}
-#endif
-
 	lssl->userPtr = options->userPtr;
 	if (options->maxFragLen < 0) {
 		/* User wants to deny a client request for changing max frag len */
@@ -278,6 +280,12 @@ int32 matrixSslNewServerSession(ssl_t **ssl, const sslKeys_t *keys,
 	if (options->truncHmac < 0) {
 		lssl->extFlags.deny_truncated_hmac = 1;
 	}
+	
+	/* Extended master secret is enabled by default.  If user sets to 1 this
+		is a flag to REQUIRE its use */
+	if (options->extendedMasterSecret > 0) {
+		lssl->extFlags.require_extended_master_secret = 1;
+	}
 
 	*ssl = lssl;
 	return MATRIXSSL_SUCCESS;
@@ -305,6 +313,8 @@ void matrixSslRegisterALPNCallback(ssl_t *ssl,
 
 #endif /* USE_SERVER_SIDE_SSL */
 
+
+
 /******************************************************************************/
 /*
 	Caller is asking for allocated buffer storage to recv data into
@@ -325,6 +335,54 @@ int32 matrixSslGetReadbuf(ssl_t *ssl, unsigned char **buf)
 	return ssl->insize - ssl->inlen;
 }
 
+/* If the required size is known, grow the buffer here so the caller doesn't
+	have to go through the REQUEST_RECV logic of matrixSslReceivedData
+
+	The return value MAY be larger than the requested size if the inbuf
+	is already larger than what was requested.
+*/
+int32 matrixSslGetReadbufOfSize(ssl_t *ssl, int32 size, unsigned char **buf)
+{
+	unsigned char *p;
+
+	if (!ssl || !buf) {
+		return PS_ARG_FAIL;
+	}
+	psAssert(ssl && ssl->insize > 0 && ssl->inbuf != NULL);
+
+	if ((ssl->insize - ssl->inlen) >= size) {
+		/* Already enough room in current buffer */
+		return matrixSslGetReadbuf(ssl, buf);
+	}
+
+	/* Going to have to grow... but do we have to realloc to save data? */
+	if (ssl->inlen == 0) {
+		/* buffer is empty anyway so can free before alloc and help keep high
+			water mark down */
+		psFree(ssl->inbuf, ssl->bufferPool);
+		ssl->inbuf = NULL;
+		if ((ssl->inbuf = psMalloc(ssl->bufferPool, size)) == NULL) {
+			ssl->insize = 0;
+			return PS_MEM_FAIL;
+		}
+		ssl->insize = size;
+		*buf = ssl->inbuf;
+		return ssl->insize;
+	} else {
+		/* realloc with: total size = current size + requested size */
+		if ((p = psRealloc(ssl->inbuf, ssl->inlen + size, ssl->bufferPool))
+				== NULL) {
+			ssl->inbuf = NULL; ssl->insize = 0; ssl->inlen = 0;
+			return PS_MEM_FAIL;
+		}
+		ssl->inbuf = p;
+		ssl->insize = ssl->inlen + size;
+		*buf = ssl->inbuf + ssl->inlen;
+		return size;
+	}
+	return PS_FAILURE; /* can't hit */
+}
+
 /******************************************************************************/
 /*
  	Caller is asking if there is any encoded, outgoing SSL data that should be
@@ -369,6 +427,9 @@ int32 matrixSslGetOutdata(ssl_t *ssl, unsigned char **buf)
 int32 matrixSslGetWritebuf(ssl_t *ssl, unsigned char **buf, uint32 requestedLen)
 {
 	uint32			requiredLen, sz, overhead;
+#ifdef USE_DTLS
+	int32			pmtu;
+#endif
 	unsigned char	*p;
 
 	if (!ssl || !buf) {
@@ -376,7 +437,6 @@ int32 matrixSslGetWritebuf(ssl_t *ssl, unsigned char **buf, uint32 requestedLen)
 	}
 	psAssert(ssl->outsize > 0 && ssl->outbuf != NULL);
 
-
 #ifdef USE_BEAST_WORKAROUND
 	/* This is a client-only feature */
 	if (!(ssl->flags & SSL_FLAGS_SERVER)) {
@@ -403,11 +463,27 @@ int32 matrixSslGetWritebuf(ssl_t *ssl, unsigned char **buf, uint32 requestedLen)
 /*
 	What is the total encoded size for a plaintext requestedLen.  The overhead
 	includes leading header as well as trailing MAC and pad
+	
+	We want to tweak the overhead an extra block to account for a
+	padding miscalculation in matrixSslGetEncodedSize.  If that call was
+	made on an exact-sized message and the user decides to use a
+	different record size than requested, we'll need to make sure
+	there is enough available room for any potential padding length.
 */
-	requiredLen = matrixSslGetEncodedSize(ssl, requestedLen);
+	requiredLen = matrixSslGetEncodedSize(ssl, requestedLen + ssl->enBlockSize);
+	
 	psAssert(requiredLen >= requestedLen);
 	overhead = requiredLen - requestedLen;
 
+#ifdef USE_DTLS
+	if (ssl->flags & SSL_FLAGS_DTLS) {
+		pmtu = matrixDtlsGetPmtu();
+		if (requiredLen > (uint32)pmtu) {
+			overhead = matrixSslGetEncodedSize(ssl, 0) + ssl->enBlockSize;
+			requiredLen = matrixSslGetEncodedSize(ssl, pmtu - overhead);
+		}
+	}
+#endif
 
 /*
 	Get current available space in outbuf
@@ -434,20 +510,13 @@ int32 matrixSslGetWritebuf(ssl_t *ssl, unsigned char **buf, uint32 requestedLen)
 			return PS_FAILURE;
 		}
 		sz = ssl->outsize - ssl->outlen;
-	} else {
-		/* If we're returning excess data from the buffer for caller use,
-			we want to tweak the overhead an extra block to account for a
-			padding miscalculation in matrixSslGetEncodedSize.  That call was
-			made on an exact-sized message, so if the user decides to use a
-			different record size than requested, we'll need to make sure
-			there is enough available room for any potential padding length. */
-		overhead += ssl->enBlockSize;
 	}
+	
 /*
 	Now that requiredLen has been confirmed/created, return number of available
 	plaintext bytes
 */
-	if (requestedLen < (uint32)ssl->maxPtFrag) {
+	if (requestedLen <= (uint32)ssl->maxPtFrag) {
 		requestedLen = sz - overhead;
 		if (requestedLen > (uint32)ssl->maxPtFrag) {
 			requestedLen = ssl->maxPtFrag;
@@ -469,15 +538,12 @@ int32 matrixSslGetWritebuf(ssl_t *ssl, unsigned char **buf, uint32 requestedLen)
 		*buf = ssl->outbuf + ssl->outlen + ssl->recordHeadLen + ssl->enBlockSize;
 		return requestedLen; /* may not be what was passed in */
 	}
-#if defined(USE_TLS_1_2) && defined(USE_AES_GCM)
-		/* GCM mode will need to save room for the nonce */
-		if ((ssl->flags & SSL_FLAGS_TLS_1_2) &&
-				(ssl->flags & SSL_FLAGS_GMAC_W)) {
-			*buf = ssl->outbuf + ssl->outlen + ssl->recordHeadLen +
-				ssl->nonceCtrLen;
-			return requestedLen; /* may not be what was passed in */
-		}
-#endif /* USE_TLS_1_2 && AES_GCM */
+	/* GCM mode will need to save room for the nonce */
+	if (ssl->flags & SSL_FLAGS_AEAD_W) {
+		*buf = ssl->outbuf + ssl->outlen + ssl->recordHeadLen +
+			AEAD_NONCE_LEN(ssl);
+		return requestedLen; /* may not be what was passed in */
+	}
 #endif /* USE_TLS_1_1 */
 
 #ifdef USE_BEAST_WORKAROUND
@@ -568,11 +634,9 @@ int32 matrixSslEncodeWritebuf(ssl_t *ssl, uint32 len)
 			(ssl->flags & SSL_FLAGS_TLS_1_1) &&	(ssl->enBlockSize > 1)) {
 		reserved += ssl->enBlockSize;
 	}
-#if defined(USE_TLS_1_2) && defined(USE_AES_GCM)
-	if ((ssl->flags & SSL_FLAGS_TLS_1_2) &&	(ssl->flags & SSL_FLAGS_GMAC_W)) {
-		reserved += ssl->nonceCtrLen;
+	if (ssl->flags & SSL_FLAGS_AEAD_W) {
+		reserved += AEAD_NONCE_LEN(ssl);
 	}
-#endif /* USE_TLS_1_2 && AES_GCM */
 #endif /* USE_TLS_1_1 */
 
 	rc = matrixSslEncode(ssl, origbuf, (ssl->outsize - ssl->outlen),
@@ -613,6 +677,14 @@ int32 matrixSslEncodeToOutdata(ssl_t *ssl, unsigned char *ptBuf, uint32 len)
 		return PS_PROTOCOL_FAIL;
 	}
 
+#ifdef USE_DTLS
+	if (ssl->flags & SSL_FLAGS_DTLS) {
+		rc = matrixSslGetEncodedSize(ssl, len);
+		if (rc > matrixDtlsGetPmtu()) {
+			return PS_LIMIT_FAIL;
+		}
+	}
+#endif
 
 	/* Fragmentation support */
 	index = 0;
@@ -648,7 +720,6 @@ int32 matrixSslEncodeToOutdata(ssl_t *ssl, unsigned char *ptBuf, uint32 len)
 	return ssl->outlen;
 }
 
-
 /******************************************************************************/
 /*
 	Helper to shrink buffers down to default size
@@ -662,7 +733,15 @@ static void revertToDefaultBufsize(ssl_t *ssl, uint16 inOrOut)
 	unsigned char	*p;
 
 	if (inOrOut == SSL_INBUF) {
+#ifdef USE_DTLS
+		if (ssl->flags & SSL_FLAGS_DTLS) {
+			defaultSize = matrixDtlsGetPmtu();
+		} else {
+			defaultSize = SSL_DEFAULT_IN_BUF_SIZE;
+		}
+#else
 		defaultSize = SSL_DEFAULT_IN_BUF_SIZE;
+#endif
 		if (ssl->insize > defaultSize && ssl->inlen < defaultSize) {
 			/* It's not fatal if we can't realloc it smaller */
 			if ((p = psRealloc(ssl->inbuf, defaultSize, ssl->bufferPool))
@@ -672,7 +751,15 @@ static void revertToDefaultBufsize(ssl_t *ssl, uint16 inOrOut)
 			}
 		}
 	} else {
+#ifdef USE_DTLS
+		if (ssl->flags & SSL_FLAGS_DTLS) {
+			defaultSize = matrixDtlsGetPmtu();
+		} else {
+			defaultSize = SSL_DEFAULT_OUT_BUF_SIZE;
+		}
+#else
 		defaultSize = SSL_DEFAULT_OUT_BUF_SIZE;
+#endif
 		if (ssl->outsize > defaultSize && ssl->outlen < defaultSize) {
 			/* It's not fatal if we can't realloc it smaller */
 			if ((p = psRealloc(ssl->outbuf, defaultSize, ssl->bufferPool))
@@ -766,10 +853,35 @@ DECODE_MORE:
 				rc = MATRIXSSL_REQUEST_RECV; /* Need to recv more handshake data */
 			}
 		} else {
+#ifdef USE_DTLS
+			rc = MATRIXSSL_REQUEST_RECV; /* Got FINISHED without CCS */
+#else
 			/* This is an error - we shouldn't get here */
+#endif
 		}
 		break;
 
+#ifdef USE_DTLS
+	case DTLS_RETRANSMIT:
+		/* Only request a resend if last record in buffer */
+		ssl->inlen -= processed;
+		if (ssl->inlen > 0) {
+			psAssert(buf > ssl->inbuf);
+/*
+			Pack ssl->inbuf so there is immediate maximum room for potential
+			outgoing data that needs to be written
+*/
+			memmove(ssl->inbuf, buf, ssl->inlen);
+			buf = ssl->inbuf;
+			goto DECODE_MORE;	/* More data in buffer to process */
+		}
+
+		/* Flight will be rebuilt when matrixDtlsGetOutdata is called while
+		outbuf is empty.  This is the return case where we are actually
+		seeing a repeat handshake message so we know something was lost in
+		flight. */
+		return MATRIXSSL_REQUEST_SEND;
+#endif
 
 	case SSL_SEND_RESPONSE:
 #ifdef ENABLE_FALSE_START
@@ -835,6 +947,9 @@ DECODE_MORE:
 		break;
 
 	case MATRIXSSL_ERROR:
+		if (decodeErr >= 0) {
+			//printf("THIS SHOULD BE A NEGATIVE VALUE?\n");
+		}
 		return decodeErr; /* Will be a negative value */
 
 	case SSL_ALERT:
@@ -876,7 +991,6 @@ DECODE_MORE:
 			/* Don't need to change inlen */
 		}
 
-
 		rc = MATRIXSSL_REQUEST_RECV;	/* Expecting more data */
 		break;
 
@@ -940,14 +1054,31 @@ DECODE_MORE:
 		/* Call user plaintext data handler */
 #ifdef USE_TLS_1_1
 		/* Been ignoring the explicit IV up to this final return point. */
+		/* NOTE: This test has been on enBlockSize for a very long time but
+			it looks like it should be on deBlockSize since this a decryption.
+			Changed and added an assert to see if these ever don't match */
+		psAssert(ssl->enBlockSize == ssl->deBlockSize);
 		if ((ssl->flags & SSL_FLAGS_READ_SECURE) &&
-				(ssl->flags & SSL_FLAGS_TLS_1_1) &&	(ssl->enBlockSize > 1)) {
-			len -= ssl->enBlockSize;
-			prevBuf += ssl->enBlockSize;
+				(ssl->flags & SSL_FLAGS_TLS_1_1) &&	(ssl->deBlockSize > 1)) {
+			len -= ssl->deBlockSize;
+			prevBuf += ssl->deBlockSize;
 		}
+		/* END enBlockSize to deBlockSize change */
 #endif /* USE_TLS_1_1 */
 		*ptbuf = prevBuf;
 		*ptlen = len;
+#ifdef USE_DTLS
+/*
+		This flag is used in conjuction with flightDone in the buffer
+		management API set to determine whether we are still in a handshake
+		state for attempting flight resends. If we are getting app data we
+		know for certain we are out of the hs states. Testing HandshakeComplete
+		is not enough because you never know if the other side got FINISHED.
+*/
+		if (ssl->flags & SSL_FLAGS_DTLS) {
+			ssl->appDataExch = 1;
+		}
+#endif
 #ifdef USE_ZLIB_COMPRESSION
 		if (ssl->compression > 0) {
 			return MATRIXSSL_APP_DATA_COMPRESSED;
@@ -999,16 +1130,14 @@ int32 matrixSslProcessedData(ssl_t *ssl, unsigned char **ptbuf, uint32 *ptlen)
 	/* Move any remaining data to the beginning of the buffer */
 	if (ssl->inlen > 0) {
 		ctlen = ssl->rec.len + ssl->recordHeadLen;
-#if defined(USE_TLS_1_2) && defined(USE_AES_GCM)
-		if (ssl->flags & SSL_FLAGS_GMAC_R) {
+		if (ssl->flags & SSL_FLAGS_AEAD_R) {
 			/* This overhead was removed from rec.len after the decryption
 				to keep buffer logic working. */
 			/* TODO: This is for checking async or not.  If async, this length
-				tweak never happen.  Need a more generic way to look for
+				tweak never happens.  Need a more generic way to look for
 				blocking or not */
-			ctlen += TLS_GCM_TAG_LEN + ssl->nonceCtrLen;
+			ctlen += AEAD_TAG_LEN(ssl) + AEAD_NONCE_LEN(ssl);
 		}
-#endif
 		memmove(ssl->inbuf, ssl->inbuf + ctlen, ssl->inlen);
 	}
 	/* Shrink inbuf to default size once inlen < default size */
@@ -1022,6 +1151,7 @@ int32 matrixSslProcessedData(ssl_t *ssl, unsigned char **ptbuf, uint32 *ptlen)
 	return MATRIXSSL_SUCCESS;
 }
 
+
 /******************************************************************************/
 /*
 	Returns < 0 on error
@@ -1050,9 +1180,6 @@ L_CLOSUREALERT:
 		rc = sslEncodeClosureAlert(ssl, &sbuf, &reqLen);
 		if (rc == SSL_FULL && newLen == 0) {
 			newLen = ssl->outlen + reqLen;
-			if (newLen > SSL_MAX_BUF_SIZE) {
-				return PS_MEM_FAIL;
-			}
 			if ((p = psRealloc(ssl->outbuf, newLen, ssl->bufferPool)) == NULL) {
 				return PS_MEM_FAIL;
 			}
@@ -1080,9 +1207,10 @@ L_CLOSUREALERT:
 
 	cipherSpec is only used by clients
  */
-int32 matrixSslEncodeRehandshake(ssl_t *ssl, sslKeys_t *keys,
+int32_t matrixSslEncodeRehandshake(ssl_t *ssl, sslKeys_t *keys,
 				int32 (*certCb)(ssl_t *ssl, psX509Cert_t *cert, int32 alert),
-				uint32 sessionOption, uint32 cipherSpec[], uint16 cipherSpecLen)
+				uint32 sessionOption,
+				const uint16_t cipherSpec[], uint8_t cipherSpecLen)
 {
 	sslBuf_t		sbuf;
 	int32			rc, i;
@@ -1090,6 +1218,10 @@ int32 matrixSslEncodeRehandshake(ssl_t *ssl, sslKeys_t *keys,
 	unsigned char	*p;
 	sslSessOpts_t	options;
 
+	/* Clear extFlags for rehandshakes */
+	ssl->extFlags.truncated_hmac = 0;
+	ssl->extFlags.sni = 0;
+	
 	if (!ssl) {
 		return PS_ARG_FAIL;
 	}
@@ -1101,6 +1233,9 @@ int32 matrixSslEncodeRehandshake(ssl_t *ssl, sslKeys_t *keys,
 	}
 	psAssert(ssl->outsize > 0 && ssl->outbuf != NULL);
 
+#ifdef DISABLE_DTLS_CLIENT_CHANGE_CIPHER_FROM_GCM_TO_GCM
+#endif /* DISABLE_DTLS_CLIENT_CHANGE_CIPHER_FROM_GCM_TO_GCM */
+
 #ifdef USE_ZLIB_COMPRESSION
 	/* Re-handshakes are not currently supported for compressed sessions. */
 	if (ssl->compression > 0) {
@@ -1129,7 +1264,7 @@ int32 matrixSslEncodeRehandshake(ssl_t *ssl, sslKeys_t *keys,
 	if (certCb != NULL) {
 		matrixSslSetSessionOption(ssl, SSL_OPTION_FULL_HANDSHAKE, NULL);
 #if defined(USE_CLIENT_AUTH) || defined(USE_CLIENT_SIDE_SSL)
-		matrixSslSetCertValidator(ssl, (sslCertCb_t)certCb);
+		matrixSslSetCertValidator(ssl, certCb);
 #endif /* USE_CLIENT_AUTH || USE_CLIENT_SIDE_SSL */
 #if defined(USE_CLIENT_AUTH) && defined(USE_SERVER_SIDE_SSL)
 /*
@@ -1160,6 +1295,15 @@ int32 matrixSslEncodeRehandshake(ssl_t *ssl, sslKeys_t *keys,
 			matrixSslSetSessionOption(ssl, SSL_OPTION_FULL_HANDSHAKE, NULL);
 		}
 	}
+#ifdef USE_DTLS
+	if (ssl->flags & SSL_FLAGS_DTLS) {
+		/* Resend epoch should be brought up-to-date with new epoch */
+		ssl->resendEpoch[0] = ssl->epoch[0];
+		ssl->resendEpoch[1] = ssl->epoch[1];
+
+		ssl->msn = ssl->resendMsn = 0;
+	}
+#endif /* USE_DTLS */
 /*
 	Options are set.  Encode the HELLO message
 */
@@ -1187,6 +1331,17 @@ L_REHANDSHAKE:
 		sbuf.buf = sbuf.start = sbuf.end = ssl->outbuf + ssl->outlen;
 		sbuf.size = ssl->outsize - ssl->outlen;
 		memset(&options, 0x0, sizeof(sslSessOpts_t));
+#ifdef USE_ECC_CIPHER_SUITE
+		options.ecFlags = ssl->ecInfo.ecFlags;
+#endif
+		/* Use extended master secret if original connection used it */
+		if (ssl->extFlags.extended_master_secret == 1) {
+			options.extendedMasterSecret = 1;
+			ssl->extFlags.extended_master_secret = 0;
+		} else {
+			options.extendedMasterSecret = -1;
+		}
+	
 		if ((rc = matrixSslEncodeClientHello(ssl, &sbuf, cipherSpec,
 				cipherSpecLen, &reqLen, NULL, &options)) < 0) {
 			if (rc == SSL_FULL && newLen == 0) {
@@ -1251,9 +1406,10 @@ void matrixSslAddRehandshakeCredits(ssl_t *ssl, int32 credits)
 }
 
 #else /* !SSL_REHANDSHAKES_ENABLED */
-int32 matrixSslEncodeRehandshake(ssl_t *ssl, sslKeys_t *keys,
+int32_t matrixSslEncodeRehandshake(ssl_t *ssl, sslKeys_t *keys,
 				int32 (*certCb)(ssl_t *ssl, psX509Cert_t *cert, int32 alert),
-				uint32 sessionOption, uint32 cipherSpec[], uint16 cipherSpecLen)
+				uint32_t sessionOption,
+				const uint16_t cipherSpec[], uint8_t cipherSpecLen)
 {
 	psTraceInfo("Rehandshaking is disabled.  matrixSslEncodeRehandshake off\n");
 	return PS_FAILURE;
@@ -1339,24 +1495,6 @@ int32 matrixSslSentData(ssl_t *ssl, uint32 bytes)
 #ifdef USE_SSL_INFORMATIONAL_TRACE
 		/* Client side resumed completion or server standard completion */
 		matrixSslPrintHSDetails(ssl);
-#endif
-#ifdef USE_UNIFIED_PKCS11
-		/* Too ugly to track DTLS client/server/normal/resumed/rehandshake cases
-			for deleting old crypto session objects for the minimum lifecycle
-			so we're just looking for any leftover cases here when we are
-			certain handshake is complete.
-
-			NOTE: It is possible this oldCrypt will still be needed if the
-			final flight is the CCS/FINISHED pair only and the application
-			has built in the smarts to resend if no app data is received. This
-			problem would manifest itself in a failed cipher init and the
-			solution would be to hold off this free until application data
-			is exchanged */
-		if (ssl->sec.oldCrypt != CK_INVALID_HANDLE &&
-				ssl->sec.oldCrypt != ssl->sec.pkcs11Ses) {
-			pkcs11CloseSession(ssl->sec.oldCrypt);
-			ssl->sec.oldCrypt = CK_INVALID_HANDLE;
-		}
 #endif
 	}
 	return rc;
@@ -1372,81 +1510,5 @@ int32 matrixSslIsSessionCompressionOn(ssl_t *ssl)
 }
 #endif
 
-
-#ifdef USE_CRL
-/*
-	Called after key load if CRL location is expected to be embedded in the CA.
-	The user callback will be invoked with the URL and the responsibility of
-	the callback is to fetch the CRL and load it to the CA with matrixSslLoadCRL
-
-	The numLoaded parameter indicates how many successful CRLs were loaded
-	if there are multiple CA files being processed here.
-
-	Return codes:
-	< 0 - Error loading a CRL.  numLoaded indicates if some success
-	0 - No errors encountered.  numLoaded could be 0 if no CRLs found
-*/
-int32 matrixSslGetCRL(sslKeys_t	*keys, int32 (*crlCb)(psPool_t *pool,
-			psX509Cert_t *CA, int append, char *url, uint32 urlLen),
-			int32 *numLoaded)
-{
-	psX509Cert_t		*CA;
-	x509GeneralName_t	*gn;
-	unsigned char		*crlURI;
-	int32				rc, crlURILen;
-
-	if (keys->CAcerts == NULL || crlCb == NULL) {
-		return PS_ARG_FAIL;
-	}
-	CA = keys->CAcerts;
-
-	*numLoaded = rc = 0;
-	while (CA) {
-		if (CA->extensions.bc.cA > 0 && CA->extensions.crlDist != NULL) {
-			gn = CA->extensions.crlDist;
-			while (gn) {
-				if (gn->id == 6) { /* Only pass on URI types */
-					crlURI = gn->data;
-					crlURILen = gn->dataLen;
-					/* Invoke user callback to go fetch and load the CRL.
-						Do not use the append flag.  The specification says
-						that multiple names must be mechanims to access the
-						same CRL. */
-					if (crlCb(keys->pool, CA, 0, (char*)crlURI,
-							crlURILen) < 0) {
-						rc = -1;
-					}
-					(*numLoaded)++; /* Callback successfully loaded this CRL */
-				} else {
-					psTraceIntInfo("Unsupported CRL distro point format %d\n",
-						gn->id);
-				}
-				gn = gn->next;
-			}
-		}
-		CA = CA->next;
-	}
-	return rc;
-}
-
-/*
-	If user already has a CRL buffer handy or call from the callback to load
-	a fetched one
-
-	Return codes:
-	< 0 - Error loading CRL
-	>= 0 - Success
-*/
-int32 matrixSslLoadCRL(psPool_t *pool, psX509Cert_t *CA, int append,
-						const char *CRLbin, int32 CRLbinLen, void *poolUserPtr)
-{
-	if (CA == NULL) {
-		return PS_ARG_FAIL;
-	}
-
-	return psX509ParseCrl(pool, CA, append, (unsigned char*)CRLbin, CRLbinLen,
-		poolUserPtr);
-}
-#endif
 /******************************************************************************/
 
