@@ -1,6 +1,6 @@
 /**
  *	@file    tls.c
- *	@version 33ef80f (HEAD, tag: MATRIXSSL-3-7-2-OPEN, tag: MATRIXSSL-3-7-2-COMM, origin/master, origin/HEAD, master)
+ *	@version 7f5c362 (HEAD -> master)
  *
  *	TLS (SSLv3.1+) specific code.
  *	http://www.faqs.org/rfcs/rfc2246.html
@@ -8,7 +8,7 @@
  *	and handshake hashing.
  */
 /*
- *	Copyright (c) 2013-2015 INSIDE Secure Corporation
+ *	Copyright (c) 2013-2016 INSIDE Secure Corporation
  *	Copyright (c) PeerSec Networks, 2002-2011
  *	All Rights Reserved
  *
@@ -22,7 +22,7 @@
  *	This General Public License does NOT permit incorporating this software
  *	into proprietary programs.  If you are unable to comply with the GPL, a
  *	commercial license for this software may be purchased from INSIDE at
- *	http://www.insidesecure.com/eng/Company/Locations
+ *	http://www.insidesecure.com/
  *
  *	This program is distributed in WITHOUT ANY WARRANTY; without even the
  *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
@@ -45,86 +45,56 @@
 #define LABEL_MASTERSEC		"master secret"
 #define LABEL_KEY_BLOCK		"key expansion"
 
-/******************************************************************************/
-/*
- *	Generates all key material.
- */
-int32 tlsDeriveKeys(ssl_t *ssl)
+#define LABEL_EXT_SIZE			22
+#define LABEL_EXT_MASTERSEC		"extended master secret"
+
+
+static int32_t genKeyBlock(ssl_t *ssl)
 {
 	unsigned char	msSeed[SSL_HS_RANDOM_SIZE * 2 + LABEL_SIZE];
 	uint32			reqKeyLen;
-
-/*
-	If this session is resumed, we want to reuse the master secret to
-	regenerate the key block with the new random values.
-*/
-	if (ssl->flags & SSL_FLAGS_RESUMED) {
-		goto skipPremaster;
-	}
-/*
-	master_secret = PRF(pre_master_secret, "master secret",
-		client_random + server_random);
-*/
-	memcpy(msSeed, LABEL_MASTERSEC, LABEL_SIZE);
-	memcpy(msSeed + LABEL_SIZE, ssl->sec.clientRandom,
-		SSL_HS_RANDOM_SIZE);
-	memcpy(msSeed + LABEL_SIZE + SSL_HS_RANDOM_SIZE,
-		ssl->sec.serverRandom, SSL_HS_RANDOM_SIZE);
-
-#ifdef USE_TLS_1_2
-	if (ssl->flags & SSL_FLAGS_TLS_1_2) {
-		prf2(ssl->sec.premaster, ssl->sec.premasterSize, msSeed,
-				(SSL_HS_RANDOM_SIZE * 2) + LABEL_SIZE, ssl->sec.masterSecret,
-				SSL_HS_MASTER_SIZE, ssl->cipher->flags);
-#ifndef USE_ONLY_TLS_1_2
-	} else {
-		prf(ssl->sec.premaster, ssl->sec.premasterSize, msSeed,
-			(SSL_HS_RANDOM_SIZE * 2) + LABEL_SIZE, ssl->sec.masterSecret,
-			SSL_HS_MASTER_SIZE);
-#endif
-	}
-#else
-	prf(ssl->sec.premaster, ssl->sec.premasterSize, msSeed,
-		(SSL_HS_RANDOM_SIZE * 2) + LABEL_SIZE, ssl->sec.masterSecret,
-		SSL_HS_MASTER_SIZE);
-#endif
-
-/*
-	 premaster is now allocated for DH reasons.  Can free here
-*/
-	psFree(ssl->sec.premaster, ssl->hsPool);
-	ssl->sec.premaster = NULL;
-	ssl->sec.premasterSize = 0;
-
-skipPremaster:
+	int32_t			rc = PS_FAIL;
+	
 	memcpy(msSeed, LABEL_KEY_BLOCK, LABEL_SIZE);
 	memcpy(msSeed + LABEL_SIZE, ssl->sec.serverRandom,
 		SSL_HS_RANDOM_SIZE);
 	memcpy(msSeed + LABEL_SIZE + SSL_HS_RANDOM_SIZE,
 		ssl->sec.clientRandom, SSL_HS_RANDOM_SIZE);
-/*
-	We must generate enough key material to fill the various keys
-*/
+
+	/* We must generate enough key material to fill the various keys */
 	reqKeyLen = 2 * ssl->cipher->macSize +
 				2 * ssl->cipher->keySize +
 				2 * ssl->cipher->ivSize;
 
+	/* Ensure there's enough room */
+	if (reqKeyLen > SSL_MAX_KEY_BLOCK_SIZE) {
+		rc = PS_MEM_FAIL;
+		goto L_RETURN;
+	}
+
 #ifdef USE_TLS_1_2
-		if (ssl->flags & SSL_FLAGS_TLS_1_2) {
-			prf2(ssl->sec.masterSecret, SSL_HS_MASTER_SIZE, msSeed,
+	if (ssl->flags & SSL_FLAGS_TLS_1_2) {
+		if ((rc = prf2(ssl->sec.masterSecret, SSL_HS_MASTER_SIZE, msSeed,
 				(SSL_HS_RANDOM_SIZE * 2) + LABEL_SIZE, ssl->sec.keyBlock,
-				reqKeyLen, ssl->cipher->flags);
-#ifndef USE_ONLY_TLS_1_2
-		} else {
-			prf(ssl->sec.masterSecret, SSL_HS_MASTER_SIZE, msSeed,
-				(SSL_HS_RANDOM_SIZE * 2) + LABEL_SIZE, ssl->sec.keyBlock,
-				reqKeyLen);
-#endif
+				reqKeyLen, ssl->cipher->flags)) < 0) {
+			goto L_RETURN;
 		}
+	}
+#ifndef USE_ONLY_TLS_1_2
+	else {
+		if ((rc = prf(ssl->sec.masterSecret, SSL_HS_MASTER_SIZE, msSeed,
+				(SSL_HS_RANDOM_SIZE * 2) + LABEL_SIZE, ssl->sec.keyBlock,
+				reqKeyLen)) < 0) {
+			goto L_RETURN;
+		}
+	}
+#endif
 #else
-	prf(ssl->sec.masterSecret, SSL_HS_MASTER_SIZE, msSeed,
-		(SSL_HS_RANDOM_SIZE * 2) + LABEL_SIZE, ssl->sec.keyBlock,
-		reqKeyLen);
+	if ((rc = prf(ssl->sec.masterSecret, SSL_HS_MASTER_SIZE, msSeed,
+			(SSL_HS_RANDOM_SIZE * 2) + LABEL_SIZE, ssl->sec.keyBlock,
+			reqKeyLen)) < 0) {
+		goto L_RETURN;
+	}
 #endif
 	if (ssl->flags & SSL_FLAGS_SERVER) {
 		ssl->sec.rMACptr = ssl->sec.keyBlock;
@@ -142,7 +112,167 @@ skipPremaster:
 		ssl->sec.rIVptr = ssl->sec.wIVptr + ssl->cipher->ivSize;
 	}
 
-	return SSL_HS_MASTER_SIZE;
+	rc = SSL_HS_MASTER_SIZE;
+
+L_RETURN:
+	memzero_s(msSeed, sizeof(msSeed));
+	if (rc < 0) {
+		memzero_s(ssl->sec.masterSecret, SSL_HS_MASTER_SIZE);
+		memzero_s(ssl->sec.keyBlock, SSL_MAX_KEY_BLOCK_SIZE);
+	}
+	return rc;
+}
+
+/******************************************************************************/
+/*
+ *	Generates all key material.
+ */
+int32_t tlsDeriveKeys(ssl_t *ssl)
+{
+	unsigned char	msSeed[SSL_HS_RANDOM_SIZE * 2 + LABEL_SIZE];
+	int32_t			rc = PS_FAIL;
+
+#ifdef USE_DTLS
+	if (ssl->flags & SSL_FLAGS_DTLS && ssl->retransmit == 1) {
+		/* The keyblock is still valid from the first pass */
+		return SSL_HS_MASTER_SIZE;
+	}
+#endif
+/*
+	If this session is resumed, we want to reuse the master secret to
+	regenerate the key block with the new random values.
+*/
+	if (ssl->flags & SSL_FLAGS_RESUMED) {
+		return genKeyBlock(ssl);
+	}
+
+/*
+	master_secret = PRF(pre_master_secret, "master secret",
+		client_random + server_random);
+*/
+	memcpy(msSeed, LABEL_MASTERSEC, LABEL_SIZE);
+	memcpy(msSeed + LABEL_SIZE, ssl->sec.clientRandom,
+		SSL_HS_RANDOM_SIZE);
+	memcpy(msSeed + LABEL_SIZE + SSL_HS_RANDOM_SIZE,
+		ssl->sec.serverRandom, SSL_HS_RANDOM_SIZE);
+
+#ifdef USE_TLS_1_2
+	if (ssl->flags & SSL_FLAGS_TLS_1_2) {
+		if ((rc = prf2(ssl->sec.premaster, ssl->sec.premasterSize, msSeed,
+				(SSL_HS_RANDOM_SIZE * 2) + LABEL_SIZE, ssl->sec.masterSecret,
+				SSL_HS_MASTER_SIZE, ssl->cipher->flags)) < 0) {
+			return rc;
+		}
+#ifndef USE_ONLY_TLS_1_2
+	} else {
+		if ((rc = prf(ssl->sec.premaster, ssl->sec.premasterSize, msSeed,
+			(SSL_HS_RANDOM_SIZE * 2) + LABEL_SIZE, ssl->sec.masterSecret,
+			SSL_HS_MASTER_SIZE)) < 0) {
+			return rc;
+		}
+#endif
+	}
+#else
+	if ((rc = prf(ssl->sec.premaster, ssl->sec.premasterSize, msSeed,
+		(SSL_HS_RANDOM_SIZE * 2) + LABEL_SIZE, ssl->sec.masterSecret,
+		SSL_HS_MASTER_SIZE)) < 0) {
+			return rc;
+		}
+#endif
+
+#ifdef USE_DTLS
+	if (ssl->flags & SSL_FLAGS_DTLS) {
+/*
+		May need premaster for retransmits.  DTLS will free this when handshake
+		is known to be complete
+*/
+		return genKeyBlock(ssl);
+	}
+#endif /* USE_DTLS */
+/*
+	 premaster is now allocated for DH reasons.  Can free here
+*/
+	psFree(ssl->sec.premaster, ssl->hsPool);
+	ssl->sec.premaster = NULL;
+	ssl->sec.premasterSize = 0;
+
+	return genKeyBlock(ssl);
+}
+
+/* Master secret generation if extended_master_secret extension is used */
+int32_t tlsExtendedDeriveKeys(ssl_t *ssl)
+{
+	unsigned char	msSeed[SHA384_HASHLEN + LABEL_EXT_SIZE];
+	unsigned char	hash[SHA384_HASHLEN];
+	uint32_t		outLen;
+	int32_t			rc = PS_FAIL;
+	
+#ifdef USE_DTLS
+	if (ssl->flags & SSL_FLAGS_DTLS && ssl->retransmit == 1) {
+		/* The keyblock is still valid from the first pass */
+		return SSL_HS_MASTER_SIZE;
+	}
+#endif
+/*
+	If this session is resumed, we should reuse the master_secret to
+	regenerate the key block with the new random values. We should not
+	be here regenerating the master_secret!
+*/
+	if (ssl->extFlags.extended_master_secret == 0 ||
+			ssl->flags & SSL_FLAGS_RESUMED) {
+		psTraceInfo("Invalid invokation of extended key derivation.\n");
+		return PS_FAIL;
+	}
+	
+	extMasterSecretSnapshotHSHash(ssl, hash, &outLen);
+/*
+	master_secret = PRF(pre_master_secret, "extended master secret",
+		session_hash);
+*/
+	memcpy(msSeed, LABEL_EXT_MASTERSEC, LABEL_EXT_SIZE);
+	memcpy(msSeed + LABEL_EXT_SIZE, hash, outLen);
+
+#ifdef USE_TLS_1_2
+	if (ssl->flags & SSL_FLAGS_TLS_1_2) {
+		if ((rc = prf2(ssl->sec.premaster, ssl->sec.premasterSize, msSeed,
+				outLen + LABEL_EXT_SIZE, ssl->sec.masterSecret,
+				SSL_HS_MASTER_SIZE, ssl->cipher->flags)) < 0) {
+			return rc;
+		}
+#ifndef USE_ONLY_TLS_1_2
+	} else {
+		if ((rc = prf(ssl->sec.premaster, ssl->sec.premasterSize, msSeed,
+			outLen + LABEL_EXT_SIZE, ssl->sec.masterSecret,
+			SSL_HS_MASTER_SIZE)) < 0) {
+			return rc;
+		}
+#endif
+	}
+#else
+	if ((rc = prf(ssl->sec.premaster, ssl->sec.premasterSize, msSeed,
+		outLen + LABEL_EXT_SIZE, ssl->sec.masterSecret,
+		SSL_HS_MASTER_SIZE)) < 0) {
+			return rc;
+		}
+#endif
+
+#ifdef USE_DTLS
+	if (ssl->flags & SSL_FLAGS_DTLS) {
+/*
+		May need premaster for retransmits.  DTLS will free this when handshake
+		is known to be complete
+*/
+		return genKeyBlock(ssl);
+	}
+#endif /* USE_DTLS */
+/*
+	 premaster is now allocated for DH reasons.  Can free here
+*/
+	psFree(ssl->sec.premaster, ssl->hsPool);
+	ssl->sec.premaster = NULL;
+	ssl->sec.premasterSize = 0;
+
+	return genKeyBlock(ssl);
 }
 
 #ifdef USE_SHA_MAC
@@ -151,14 +281,22 @@ skipPremaster:
 /*
 	TLS sha1 HMAC generate/verify
 */
-int32 tlsHMACSha1(ssl_t *ssl, int32 mode, unsigned char type,
+int32_t tlsHMACSha1(ssl_t *ssl, int32 mode, unsigned char type,
 			unsigned char *data, uint32 len, unsigned char *mac)
 {
-	psHmacContext_t		ctx;
+#ifndef USE_HMAC_TLS
+	psHmacSha1_t		ctx;
+#endif
 	unsigned char		*key, *seq;
 	unsigned char		majVer, minVer, tmp[5];
 	int32				i;
-
+#ifdef USE_DTLS
+	unsigned char		dtls_seq[8];
+#endif /* USE_DTLS */
+#ifdef USE_HMAC_TLS
+	uint32				alt_len;
+#endif /* USE_HMAC_TLS */
+	
 	majVer = ssl->majVer;
 	minVer = ssl->minVer;
 
@@ -169,39 +307,82 @@ int32 tlsHMACSha1(ssl_t *ssl, int32 mode, unsigned char type,
 		key = ssl->sec.readMAC;
 		seq = ssl->sec.remSeq;
 	}
+	
+	/* Sanity */
+	if (key == NULL) {
+		return PS_FAILURE;
+	}
 
-	psHmacSha1Init(&ctx, key, SHA1_HASH_SIZE);
-		psHmacSha1Update(&ctx, seq, 8);
-		for (i = 7; i >= 0; i--) {
-			seq[i]++;
-			if (seq[i] != 0) {
-				break;
-			}
+#ifdef USE_DTLS
+	if (ssl->flags & SSL_FLAGS_DTLS) {
+		if (mode == HMAC_CREATE) {
+			seq = dtls_seq;
+			memcpy(dtls_seq, ssl->epoch, 2);
+			memcpy(dtls_seq + 2, ssl->rsn, 6);
+		} else { /* HMAC_VERIFY */
+			seq = dtls_seq;
+			memcpy(dtls_seq, ssl->rec.epoch, 2);
+			memcpy(dtls_seq + 2, ssl->rec.rsn, 6);
 		}
+	}
+#endif /* USE_DTLS */
 
 	tmp[0] = type;
 	tmp[1] = majVer;
 	tmp[2] = minVer;
 	tmp[3] = (len & 0xFF00) >> 8;
 	tmp[4] = len & 0xFF;
+#ifdef USE_HMAC_TLS
+#ifdef USE_HMAC_TLS_LUCKY13_COUNTERMEASURE
+	alt_len = mode == HMAC_CREATE ? len : ssl->rec.len;
+#else
+	alt_len = len;
+#endif
+	(void)psHmacSha1Tls(key, SHA1_HASH_SIZE,
+						seq, 8,
+						tmp, 5,
+						data, len, alt_len,
+						mac);
+#else
+	if (psHmacSha1Init(&ctx, key, SHA1_HASH_SIZE) < 0) {
+		return PS_FAIL;
+	}
+	psHmacSha1Update(&ctx, seq, 8);
 	psHmacSha1Update(&ctx, tmp, 5);
 	psHmacSha1Update(&ctx, data, len);
-	return psHmacSha1Final(&ctx, mac);
+	psHmacSha1Final(&ctx, mac);
+#endif
+	/* Update seq (only for normal TLS) */
+	for (i = 7; i >= 0; i--) {
+		seq[i]++;
+		if (seq[i] != 0) {
+			break;
+		}
+	}
+	return PS_SUCCESS;
 }
 #endif /* USE_SHA1 */
 
-#ifdef USE_SHA256
+#if defined(USE_HMAC_SHA256) || defined(USE_HMAC_SHA384)
 /******************************************************************************/
 /*
-	TLS sha2 HMAC generate/verify
+	TLS sha256/sha384 HMAC generate/verify
 */
-int32 tlsHMACSha2(ssl_t *ssl, int32 mode, unsigned char type,
+int32_t tlsHMACSha2(ssl_t *ssl, int32 mode, unsigned char type,
 			unsigned char *data, uint32 len, unsigned char *mac, int32 hashLen)
 {
-	psHmacContext_t		ctx;
+#ifndef USE_HMAC_TLS
+	psHmac_t			ctx;
+#endif
 	unsigned char		*key, *seq;
 	unsigned char		majVer, minVer, tmp[5];
 	int32				i;
+#ifdef USE_DTLS
+	unsigned char		dtls_seq[8];
+#endif /* USE_DTLS */
+#ifdef USE_HMAC_TLS
+	uint32				alt_len;
+#endif /* USE_HMAC_TLS */
 
 	majVer = ssl->majVer;
 	minVer = ssl->minVer;
@@ -213,26 +394,72 @@ int32 tlsHMACSha2(ssl_t *ssl, int32 mode, unsigned char type,
 		key = ssl->sec.readMAC;
 		seq = ssl->sec.remSeq;
 	}
+	/* Sanity */
+	if (key == NULL) {
+		return PS_FAILURE;
+	}
 
-	psHmacSha2Init(&ctx, key, hashLen, hashLen);
-		psHmacSha2Update(&ctx, seq, 8, hashLen);
-		for (i = 7; i >= 0; i--) {
-			seq[i]++;
-			if (seq[i] != 0) {
-				break;
-			}
+#ifdef USE_DTLS
+	if (ssl->flags & SSL_FLAGS_DTLS) {
+		if (mode == HMAC_CREATE) {
+			seq = dtls_seq;
+			memcpy(dtls_seq, ssl->epoch, 2);
+			memcpy(dtls_seq + 2, ssl->rsn, 6);
+		} else { /* HMAC_VERIFY */
+			seq = dtls_seq;
+			memcpy(dtls_seq, ssl->rec.epoch, 2);
+			memcpy(dtls_seq + 2, ssl->rec.rsn, 6);
 		}
+	}
+#endif /* USE_DTLS */
 
 	tmp[0] = type;
 	tmp[1] = majVer;
 	tmp[2] = minVer;
 	tmp[3] = (len & 0xFF00) >> 8;
 	tmp[4] = len & 0xFF;
-	psHmacSha2Update(&ctx, tmp, 5, hashLen);
-	psHmacSha2Update(&ctx, data, len, hashLen);
-	return psHmacSha2Final(&ctx, mac, hashLen);
+
+#ifdef USE_HMAC_TLS
+#ifdef USE_HMAC_TLS_LUCKY13_COUNTERMEASURE
+	alt_len = mode == HMAC_CREATE ? len : ssl->rec.len;
+#else
+	alt_len = len;
+#endif
+	(void)psHmacSha2Tls(key, hashLen,
+						seq, 8,
+						tmp, 5,
+						data, len, alt_len,
+						mac, hashLen);
+#else
+	switch(hashLen) {
+	case SHA256_HASHLEN:
+		if (psHmacInit(&ctx, HMAC_SHA256, key, hashLen) < 0) {
+			return PS_FAIL;
+		}
+		break;
+	case SHA384_HASHLEN:
+		if (psHmacInit(&ctx, HMAC_SHA384, key, hashLen) < 0) {
+			return PS_FAIL;
+		}
+		break;
+	default:
+		return PS_FAIL;
+	}
+	psHmacUpdate(&ctx, seq, 8);
+	psHmacUpdate(&ctx, tmp, 5);
+	psHmacUpdate(&ctx, data, len);
+	psHmacFinal(&ctx, mac);
+#endif
+	/* Update seq (only for normal TLS) */
+	for (i = 7; i >= 0; i--) {
+		seq[i]++;
+		if (seq[i] != 0) {
+			break;
+		}
+	}
+	return PS_SUCCESS;
 }
-#endif /* USE_SHA256 */
+#endif /* USE_SHA256 || USE_SHA384 */
 #endif /* USE_SHA_MAC */
 
 #ifdef USE_MD5
@@ -241,10 +468,10 @@ int32 tlsHMACSha2(ssl_t *ssl, int32 mode, unsigned char type,
 /*
 	TLS MD5 HMAC generate/verify
 */
-int32 tlsHMACMd5(ssl_t *ssl, int32 mode, unsigned char type,
+int32_t tlsHMACMd5(ssl_t *ssl, int32 mode, unsigned char type,
 				 unsigned char *data, uint32 len, unsigned char *mac)
 {
-	psHmacContext_t		ctx;
+	psHmacMd5_t			ctx;
 	unsigned char		*key, *seq;
 	unsigned char		majVer, minVer, tmp[5];
 	int32				i;
@@ -259,8 +486,25 @@ int32 tlsHMACMd5(ssl_t *ssl, int32 mode, unsigned char type,
 		key = ssl->sec.readMAC;
 		seq = ssl->sec.remSeq;
 	}
+	/* Sanity */
+	if (key == NULL) {
+		return PS_FAILURE;
+	}
 
-	psHmacMd5Init(&ctx, key, MD5_HASH_SIZE);
+	if (psHmacMd5Init(&ctx, key, MD5_HASH_SIZE) < 0) {
+		return PS_FAIL;
+	}
+#ifdef USE_DTLS
+	if (ssl->flags & SSL_FLAGS_DTLS) {
+		if (mode == HMAC_CREATE) {
+			psHmacMd5Update(&ctx, ssl->epoch, 2);
+			psHmacMd5Update(&ctx, ssl->rsn, 6);
+		} else { /* HMAC_VERIFY */
+			psHmacMd5Update(&ctx, ssl->rec.epoch, 2);
+			psHmacMd5Update(&ctx, ssl->rec.rsn, 6);
+		}
+	} else {
+#endif /* USE_DTLS */
 		psHmacMd5Update(&ctx, seq, 8);
 		for (i = 7; i >= 0; i--) {
 			seq[i]++;
@@ -268,6 +512,9 @@ int32 tlsHMACMd5(ssl_t *ssl, int32 mode, unsigned char type,
 				break;
 			}
 		}
+#ifdef USE_DTLS
+	}
+#endif /* USE_DTLS */
 
 	tmp[0] = type;
 	tmp[1] = majVer;
@@ -276,7 +523,9 @@ int32 tlsHMACMd5(ssl_t *ssl, int32 mode, unsigned char type,
 	tmp[4] = len & 0xFF;
 	psHmacMd5Update(&ctx, tmp, 5);
 	psHmacMd5Update(&ctx, data, len);
-	return psHmacMd5Final(&ctx, mac);
+	psHmacMd5Final(&ctx, mac);
+
+	return PS_SUCCESS;
 }
 #endif /* USE_MD5_MAC */
 #endif /* USE_MD5 */
@@ -310,7 +559,7 @@ int32 sslCreateKeys(ssl_t *ssl)
 */
 int32 sslActivateReadCipher(ssl_t *ssl)
 {
-
+	
 	ssl->decrypt = ssl->cipher->decrypt;
 	ssl->verifyMac = ssl->cipher->verifyMac;
 	ssl->nativeDeMacSize = ssl->cipher->macSize;
@@ -331,16 +580,27 @@ int32 sslActivateReadCipher(ssl_t *ssl)
 	memset(ssl->sec.remSeq, 0x0, sizeof(ssl->sec.remSeq));
 
 	if (ssl->cipher->ident != SSL_NULL_WITH_NULL_NULL) {
+		/* Sanity */
+		if (ssl->sec.rMACptr == NULL) {
+			psTraceInfo("sslActivateReadCipher sanity fail\n");
+			return PS_FAILURE;
+		}
 		ssl->flags |= SSL_FLAGS_READ_SECURE;
 
 #ifdef USE_TLS_1_2
 		if (ssl->deMacSize == 0) {
-			/* Need a concept for GMAC read and write start times for the
+			/* Need a concept for AEAD read and write start times for the
 				cases surrounding changeCipherSpec if moving from one suite
 				to another */
-			ssl->flags |= SSL_FLAGS_GMAC_R;
+			ssl->flags |= SSL_FLAGS_AEAD_R;
+			if (ssl->cipher->flags & CRYPTO_FLAGS_CHACHA) {
+				ssl->flags &= ~SSL_FLAGS_NONCE_R;
+			} else {
+				ssl->flags |= SSL_FLAGS_NONCE_R;
+			}
 		} else {
-			ssl->flags &= ~SSL_FLAGS_GMAC_R;
+			ssl->flags &= ~SSL_FLAGS_AEAD_R;
+			ssl->flags &= ~SSL_FLAGS_NONCE_R;
 		}
 #endif
 /*
@@ -366,6 +626,22 @@ int32 sslActivateReadCipher(ssl_t *ssl)
 
 int32 sslActivateWriteCipher(ssl_t *ssl)
 {
+#ifdef USE_DTLS
+	if (ssl->flags & SSL_FLAGS_DTLS) {
+		if (ssl->retransmit == 0) {
+			ssl->oencrypt = ssl->encrypt;
+			ssl->ogenerateMac = ssl->generateMac;
+			ssl->oenMacSize = ssl->enMacSize;
+			ssl->oenNativeHmacSize = ssl->nativeEnMacSize;
+			ssl->oenBlockSize = ssl->enBlockSize;
+			ssl->oenIvSize = ssl->enIvSize;
+			memcpy(ssl->owriteMAC, ssl->sec.writeMAC, ssl->enMacSize);
+			memcpy(&ssl->oencryptCtx, &ssl->sec.encryptCtx,
+				   sizeof(psCipherContext_t));
+			memcpy(ssl->owriteIV, ssl->sec.writeIV, ssl->cipher->ivSize);
+		}
+	}
+#endif /* USE_DTLS */
 
 	ssl->encrypt = ssl->cipher->encrypt;
 	ssl->generateMac = ssl->cipher->generateMac;
@@ -390,12 +666,18 @@ int32 sslActivateWriteCipher(ssl_t *ssl)
 
 #ifdef USE_TLS_1_2
 		if (ssl->enMacSize == 0) {
-			/* Need a concept for GMAC read and write start times for the
+			/* Need a concept for AEAD read and write start times for the
 				cases surrounding changeCipherSpec if moving from one suite
 				to another */
-			ssl->flags |= SSL_FLAGS_GMAC_W;
+			ssl->flags |= SSL_FLAGS_AEAD_W;
+			if (ssl->cipher->flags & CRYPTO_FLAGS_CHACHA) {
+				ssl->flags &= ~SSL_FLAGS_NONCE_W;
+			} else {
+				ssl->flags |= SSL_FLAGS_NONCE_W;
+			}
 		} else {
-			ssl->flags &= ~SSL_FLAGS_GMAC_W;
+			ssl->flags &= ~SSL_FLAGS_AEAD_W;
+			ssl->flags &= ~SSL_FLAGS_NONCE_W;
 		}
 #endif
 
@@ -425,14 +707,13 @@ int32 sslActivateWriteCipher(ssl_t *ssl)
 #ifdef USE_CLIENT_SIDE_SSL
 /******************************************************************************/
 /*
-	Allocate a tlsExtenstion_t structure
+	Allocate a tlsExtension_t structure
 */
 int32 matrixSslNewHelloExtension(tlsExtension_t **extension, void *userPoolPtr)
 {
 	psPool_t		*pool = NULL;
 	tlsExtension_t	*ext;
 
-
 	ext = psMalloc(pool, sizeof(tlsExtension_t));
 	if (ext == NULL) {
 		return PS_MEM_FAIL;
@@ -446,39 +727,31 @@ int32 matrixSslNewHelloExtension(tlsExtension_t **extension, void *userPoolPtr)
 
 /******************************************************************************/
 /*
-	Free a tlsExtenstion_t structure and any extensions that have been loaded
+	Free a tlsExtension_t structure and any extensions that have been loaded
 */
 void matrixSslDeleteHelloExtension(tlsExtension_t *extension)
 {
-	psPool_t		*pool;
-	tlsExtension_t	*next, *ext;
+	tlsExtension_t		*next, *ext;
 
 	if (extension == NULL) {
 		return;
 	}
 	ext = extension;
-	pool = ext->pool;
-
-/*
-	Free first one
-*/
+	/* Free first one */
 	if (ext->extData) {
-		psFree(ext->extData, pool);
+		psFree(ext->extData, ext->pool);
 	}
 	next = ext->next;
-	psFree(ext, pool);
-/*
-	Free others
-*/
+	psFree(ext, ext->pool);
+	/* Free others */
 	while (next) {
 		ext = next;
 		next = ext->next;
 		if (ext->extData) {
-			psFree(ext->extData, pool);
+			psFree(ext->extData, ext->pool);
 		}
-		psFree(ext, pool);
+		psFree(ext, ext->pool);
 	}
-
 	return;
 }
 
